/*
 * chipsound.h - ChipSound realtime scriptable sound engine
 *
 * Copyright (C) 2010-2012 David Olofson <david@olofson.net>
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from the
 * use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software
 *    in a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 * 3. This notice may not be removed or altered from any source distribution.
 */

#ifndef CHIPSOUND_H
#define CHIPSOUND_H

/* Versioning */
#define	CS_MAKE_VERSION(major, minor, micro, build)	\
		(((major) << 24) | ((minor) << 16) | ((micro) << 8) | (build))
#define	CS_MAJOR(ver)	(((ver) >> 24) & 0xff)
#define	CS_MINOR(ver)	(((ver) >> 16) & 0xff)
#define	CS_MICRO(ver)	(((ver) >> 8) & 0xff)
#define	CS_BUILD(ver)	((ver) & 0xff)

/* Current version */
#define	CS_VERSION	CS_MAKE_VERSION(0,1,0,0)

/* Default reference frequence for linear pitch 0.0; "middle C" */
#define	CS_MIDDLEC	261.626f

/* Number of VM registers */
#define	CS_REGISTERS	32

/* Group and voice ID ranges */
#define	CS_MAXGIDS	256
#define	CS_MAXVIDS	65536

/* Maximum number of arguments to a VM program or function */
#define	CS_MAXARGS	8

/* Maximum number of entry points a VM program can have. (EP 0 is "main()".) */
#define	CS_MAXEPS	8

/* Number of waveform mipmap levels */
#define	CS_MIPLEVELS	10

/* Maximum number of sample frames to process at a time */
#define	CS_MAXFRAG	64

/* Maximum number of audio channels supported */
#define	CS_MAXCHANNELS	8

typedef unsigned char CS_gid;	/* Group ID; gid 0 is the hardwired "root" */
typedef unsigned short CS_vid;	/* Voice ID; 0 is "no ID", ie detached */

/* Object types */
typedef enum CS_otypes
{
	CS_TBANK = 0,	CS_TWAVE,	CS_TPROGRAM,	CS_TUNIT
} CS_otypes;


/*---------------------------------------------------------
	Error handling
---------------------------------------------------------*/

typedef enum CS_errors
{
/*0*/	CS_OK = 0,	/* Ok - no error! */
	CS_END,		/* VM program ended normally (not an error) */
	CS_MEMORY,	/* Out of memory */
	CS_BUFOVERFLOW,	/* Buffer overflow! */
	CS_DIVBYZERO,	/* Division by zero (compile or run time!) */
	CS_OPEN,	/* Error opening file */
	CS_DEVICEOPEN,	/* Error opening device */
	CS_BADFORMAT,	/* Bad file, sample or device I/O format */
/*8*/	CS_READ,	/* Error reading file */
	CS_OVERLOAD,	/* VM overload; too many instructions back-to-back */
	CS_VOICEALLOC,	/* Could not allocate voice */
	CS_NOTRUNNING,	/* ChipSound not initialized or running */
	CS_NOTFOUND,	/* Object not found */
	CS_BADBANK,	/* Invalid bank handle */
	CS_BADWAVE,	/* Invalid waveform handle */
	CS_BADPROGRAM,	/* Bad VM program handle */
/*16*/	CS_BADENTRY,	/* Bad VM program entry point */
	CS_BADGROUP,	/* Group does not exist, or cannot be used here */
	CS_BADVOICE,	/* Voice does not exist, or bad voice id */
	CS_MANYARGS,	/* Too many arguments to VM program */
	CS_EXPORTDECL,	/* Export already declared */
	CS_EXPVALREG,	/* Expected value or register */
	CS_EXPEOS,	/* Expected end of statement */
	CS_EXPCLOSE,	/* Expected closing brace */
/*24*/	CS_EXPLABEL,	/* Expected label */
	CS_EXPPROGRAM,	/* Expected program */
	CS_EXPNAME,	/* Expected name */
	CS_EXPBODY,	/* Expected program/handler body */
	CS_EXPOP,	/* Expected operator */
	CS_BADLABEL,	/* Bad label name */
	CS_SYMBOLDEF,	/* Symbol already defined */
	CS_NEXPNAME,	/* Name not expected here */
/*32*/	CS_EXPVARIABLE,	/* Expected register variable */
	CS_EXPVALUE,	/* Expected value */
	CS_NEXPVALUE,	/* Value not expected here */
	CS_NEXPHANDLE,	/* Handle not expected here */
	CS_BADVALUE,	/* Bad value */
	CS_OVERFLOW,	/* Value does not fit in numeric type */
	CS_UNDERFLOW,	/* Value too small; would truncate to zero */
	CS_NEXPTOKEN,	/* Unexpected token */
/*40*/	CS_NEXPEOF,	/* Unexpected end of file */
	CS_BADJUMP,	/* Illegal jump target position */
	CS_INFLOOP,	/* Jump would cause infinite loop */
	CS_BADOPCODE,	/* Invalid VM opcode */
	CS_BADREGISTER,	/* Invalid VM register index */
	CS_BADREG2,	/* Invalid VM register index, second argument */
	CS_BADIMMARG,	/* Immediate argument out of range */
	CS_OUTOFREGS,	/* Out of VM registers! */
/*48*/	CS_NOPROGHERE,	/* Program cannot be declared here */
	CS_NOMSGHERE,	/* Message cannot be declared here */
	CS_BADVARDECL,	/* Variable cannot be declared here */
	CS_DECLNOINIT,	/* Declared variable not initialized */
	CS_NOTUNARY,	/* Not a unary operator */
	CS_NOCODE,	/* Code not allowed here */
	CS_NOTIMING,	/* Timing instructions not allowed here */
	CS_NORUN,	/* Cannot run program from here */
/*56*/	CS_UNDEFSYM,	/* Undefined symbols in program */
	CS_NEXPELSE,	/* 'else' not applicable to this construct */
	CS_NOWAKEFORCE,	/* 'wake' and 'force' not applicable here */
	CS_EXPUNIT,	/* Expected unit */
	CS_INTERNAL	/* Internal errors from here and up */
} CS_errors;

const char *cs_ErrorString(unsigned errorcode);


/*---------------------------------------------------------
	Versioning
---------------------------------------------------------*/

/* Return version of the linked ChipSound library. */
static inline unsigned cs_HeaderVersion(void)
{
	return CS_VERSION;
}

unsigned cs_LinkedVersion(void);


/*---------------------------------------------------------
	Open/close
---------------------------------------------------------*/

typedef enum CS_initflags
{
	CS_EXPORTALL =	0x00000001,	/* Export all programs! (Debug/tool) */
	CS_TIMESTAMP =	0x00000002,	/* Enable the cs_Now()/cs_Wait() API */
	CS_MIDI =	0x00010000,	/* Use MIDI input */
} CS_initflags;

CS_errors cs_Open(int samplerate, int buffer, int channels, int flags);
void cs_Close(void);


/*---------------------------------------------------------
	Loading/unloading sounds
---------------------------------------------------------*/

/*
 * Create a new, empty bank. 'name' is the import name for .CSA files to use;
 * NULL results in a unique name being generated automatically.
 */
int cs_NewBank(const char *name);

/*
 * Load .CSL file 'fn' or null terminated string 'code' as a bank.
 *
 * Returns the handle of the resulting bank, or if the operation fails, a
 * negative error code. (Use (-result) to get the CS_errors code.)
 */
int cs_LoadString(const char *code, const char *name);
int cs_Load(const char *fn);

/*
 * Unload bank 'handle'; -1 for "all banks". Any voices playing waves or running
 * programs in unloaded banks are killed. The built-in "root" bank cannot be
 * freed.
 */
void cs_FreeBank(int handle);


/*---------------------------------------------------------
	Waveform management
---------------------------------------------------------*/

typedef enum CS_sampleformats
{
	CS_I8 = 0,	/* [-128, 127] */
	CS_I16,		/* [-32768, 32767] */
	CS_I32,		/* [-2147483648, 2147483647] */
	CS_F32		/* [-1.0f, 1.0f] (Range irrelevant with CS_NORMALIZE) */
} CS_sampleformats;

typedef enum CS_waveflags
{
	CS_LOOPED =	0x00000001,	/* Waveform is looped */
	CS_NORMALIZE =	0x00010000,	/* Normalize waveform amplitude */
	CS_WRAP =	0x00020000,	/* Wrap/mix until finished */
	CS_XFADE =	0x00040000,	/* Crossfade to make seamless */
	CS_REVMIX =	0x00080000	/* Mix in reversed to make seemless */
} CS_waveflags;

/*
 * Upload a waveform for use by wavetable oscillators. Returns the handle of the
 * waveform, or a negative error code.
 *
 * 'bank' is the bank this waveform should be added to. 0 is the "root" bank
 * with the built-in waveforms.
 *
 * 'name' is the symbolic name by which the waveform will be known to CSASM
 *  code.
 *
 * 'length' is the length of the waveform (mipmap level 0) in sample frames.
 *
 * 'period' is the number of sample frames in one period of the waveform's
 * fundamental frequency. (Used for pitch calculations.)
 *
 * 'flags' is a set of |'ed together flags from CS_waveflags;
 *	CS_LOOPED	Wave is looped. (Affects pre-processing and playback!)
TODO:	CS_NORMALIZE	Normalize ("maximize") amplitude in the conversion.
TODO:	CS_XFADE	Crossfade mix a copy offset by half the loop length.
TODO:	CS_REVMIX	Mix wave with a reversed version of itself.
 * CS_XFADE and CS_REVMIX are intended for looped waves, although they (sort of)
 * work on one-shot waves as well.
 *
 * 'fmt' is the sample format code for the 'data'.
 *
 * 'size' is the size in BYTES of 'data'.
 */
int cs_UploadWave(unsigned bank, const char *name, unsigned period, int flags,
		CS_sampleformats fmt, const void *data, unsigned size);

/*TODO: Render CSL program into wave. */
int cs_RenderWave(unsigned bank, const char *name, unsigned period, int flags,
		unsigned samplerate, unsigned length,
		unsigned program, unsigned argc, int *argv);

/*TODO: Render CSL program, provided as source code string, into wave. */
int cs_RenderString(unsigned bank, const char *name, unsigned period, int flags,
		unsigned samplerate, unsigned length,
		const char *source);


/*---------------------------------------------------------
	Finding exports
---------------------------------------------------------*/

/*
 * Find export 'name' in 'bank' and return it's value; a code position for use
 * with cs_Start*(). If 'bank' is -1, symbols from all banks are considered.
 *
 * Returns a negative error code if no symbol was found.
 */
int cs_Find(int bank, const char *name);

/* Get handle of export by index 'i'. Returns -1 if 'i' is out of range. */
int cs_Get(int bank, unsigned i);

/* Get name of export by index 'i'. Returns NULL if 'i' is out of range. */
const char *cs_GetName(int bank, unsigned i);

/* Return type of object with 'handle', or -1 if 'handle' is out of range. */
int cs_GetType(unsigned handle);

/* Return name of 'type'. */
const char *cs_TypeName(int type);


/*---------------------------------------------------------
	Playing and controlling
---------------------------------------------------------*/

/*
 * Schedule subsequent commands to be executed as soon as possible with constant
 * latency.
 */
void cs_Now(void);

/*
 * Bump the timestamp for subsequent commands. (milliseconds)
 */
void cs_Wait(float dt);

/*
 * Create new voice group 'gid' under group 'parent'. ID 0 is a hardwired "root"
 * group, and is not available for user groups.
 */
CS_errors cs_NewGroup(CS_gid parent, CS_gid gid);

/*
 * Send a message to handler 'h' of the group driver program of group 'gid'.
 *
 * Currently implemented messages in the standard group driver:
 *	1(NV)		Set group volume to NV.
 */
CS_errors cs_SendGroupa(CS_gid gid, unsigned h, unsigned argc, int *argv);
#define cs_SendGroup(gid, h, args...)					\
	({								\
		float fa[] = { args };					\
		int i, ia[sizeof(fa) / sizeof(float)];			\
		for(i = 0; i < sizeof(ia) / sizeof(int); ++i)		\
			ia[i] = fa[i] * 65536.0f;			\
		cs_SendGroupa(gid, h, sizeof(ia) / sizeof(int), ia);	\
	})

/*
 * Detaches group 'gid'. After any voices and subgroups of the group have
 * finished, the group is removed. This works on group 0 ("root") as well, but
 * that group will not actually be removed.
 */
CS_errors cs_DetachGroup(CS_gid gid);

/*
 * Start a voice in group 'gid' and point it at 'program'. 'vid' is the ID of
 * the new voice, or 0 to start the voice detached.
 *
 * The cs_Starta() versions expect arrays of 16:16 fixed point values.
 *
 * The cs_Start() macro converts and passes any number of floating point
 * arguments.
 *
 * NOTE:
 *	Although detached voices cannot be addressed directly, they still
 *	recieve messages sent to all voices in the group!
 */
CS_errors cs_Starta(CS_gid gid, CS_vid vid, unsigned program,
		unsigned argc, int *argv);
#define cs_Start(gid, vid, prg, args...)				\
	({								\
		float fa[] = { args };					\
		int i, ia[sizeof(fa) / sizeof(float)];			\
		for(i = 0; i < sizeof(ia) / sizeof(int); ++i)		\
			ia[i] = fa[i] * 65536.0f;			\
		cs_Starta(gid, vid, prg, sizeof(ia) / sizeof(int), ia);	\
	})

/*
 * Send a message to handler 'h' of the program running on voice 'vid' of group
 * 'gid'. If 'vid' is 0, the message is sent to all voices in the group.
 */
CS_errors cs_Senda(CS_gid gid, CS_vid vid, unsigned h,
		unsigned argc, int *argv);
#define cs_Send(gid, vid, h, args...)					\
	({								\
		float fa[] = { args };					\
		int i, ia[sizeof(fa) / sizeof(float)];			\
		for(i = 0; i < sizeof(ia) / sizeof(int); ++i)		\
			ia[i] = fa[i] * 65536.0f;			\
		cs_Senda(gid, vid, h, sizeof(ia) / sizeof(int), ia);	\
	})

/*
 * Detach voice 'vid' of group 'gid'. If 'vid' is 0, all voices in the group are
 * detached. Detached voices are expected to finish any programs in progress and
 * then terminate.
 *
 * NOTE:
 *	Detached voices still recieve messages sent to all voices in the group!
 */
CS_errors cs_Detach(CS_gid gid, CS_vid vid);

/*
 * Instantly stop and detach voice 'vid' of group 'gid'. If 'vid' is 0, the
 * operation affects all voices in the group.
 */
CS_errors cs_Kill(CS_gid gid, CS_vid vid);


/*---------------------------------------------------------
	Plugin interface
---------------------------------------------------------*/

typedef CS_errors (*CS_insert_cb)(int **buffers, unsigned nbuffers,
		unsigned frames, void *userdata);

/*
 * Add 'callback' as an "insert" plugin to 'group'. The plugin will run after
 * the subvoices and before the output stage of 'group', allowing it to capture,
 * process and/or inject audio. 'userdata' will be passed to 'callback'.
 */
CS_errors cs_Insert(CS_gid gid, CS_insert_cb callback, void *userdata);


/*---------------------------------------------------------
	Utilities
---------------------------------------------------------*/

/* Convert Hz to linear pitch */
float cs_F2P(float f);

/* Return pseudo-random number in the range [0, max[ */
float cs_Rand(float max);


/*---------------------------------------------------------
	Object property interface
---------------------------------------------------------*/

typedef enum CS_properties
{
	/* Voice control registers */
	CS_PWAVE =	0x00000000,	/* 'w' */
	CS_PPITCH,			/* 'p' */
	CS_PSPITCH,			/* 'sp' */
	CS_PAMPLITUDE,			/* 'a' */
	CS_PSAMPLITUDE,			/* 'sa' */
	CS_PTICK,			/* 'tick' */

	/* Global settings and profiling information */
	CS_PSAMPLERATE=	0x00010000,	/* Audio I/O sample rate */
	CS_PBUFFER,			/* Audio I/O buffer size */
	CS_PACTIVEVOICES,		/* Number of active voices */
	CS_PFREEVOICES,			/* Number of voices in pool */
	CS_PTOTALVOICES,		/* Number of voices in total */
	CS_PCPULOADAVG,			/* Average DSP CPU load (%) */
	CS_PCPULOADMAX,			/* Peak DSP CPU load (%) */
	CS_PCPUTIMEAVG,			/* Average buffer processing time (ms) */
	CS_PCPUTIMEMAX,			/* Peak buffer processing time (ms) */
	CS_PINSTRUCTIONS,		/* VM instructions executed */
	CS_PEXPORTALL,			/* Export all programs! (Debug) */

	/* General properties */
	CS_PCHANNELS =	0x00020000,	/* Number of channels */
	CS_PFLAGS,			/* Flags */

	/* Wave properties */
	CS_PLOOPED =	0x00030000	/* Waveform is looped */
} CS_properties;

int cs_GetGlobalProp(CS_properties prop);
CS_errors cs_SetGlobalProp(CS_properties prop, int v);

int cs_GetObjectProp(unsigned handle, CS_properties prop);

int cs_GetGroupProp(CS_gid gid, CS_properties prop);

int cs_GetVoiceProp(CS_gid gid, CS_vid vid, CS_properties prop);

#endif /* CHIPSOUND_H */
