/*
 * chipsound.c - ChipSound core - internal realtime stuff, mostly
 *
 * Copyright (C) 2010-2012 David Olofson <david@olofson.net>
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from the
 * use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software
 *    in a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 * 3. This notice may not be removed or altered from any source distribution.
 */

#include <stdlib.h>
#include <math.h>
#include <stdint.h>
#include "csinternals.h"

#define	TIMESTAMPWARNINGS	0

CS_state cs_globalstate;	/* Global engine state */


/*---------------------------------------------------------
	Oscillators
---------------------------------------------------------*/

void cso_Off(CS_oscillator *o, int *buffer, unsigned frames)
{
	cs_PrepareRamp(&o->a, frames);
	cs_RunRamp(&o->a, frames);
#if 0
	/* Update, so the silent oscillator can still be used as a timer! */
	o->phase += o->dphase * frames;
#endif
}


void cso_Noise(CS_oscillator *o, int *buffer, unsigned frames)
{
	unsigned s;
	cs_PrepareRamp(&o->a, frames);
	for(s = 0; s < frames; ++s)
	{
		unsigned nph = o->phase + o->dphase;
		if((o->dphase >= 32768) || ((nph ^ o->phase) >> 15))
			o->noise = cs_Noise(&cs_globalstate) - 32767;
		o->phase = nph;
		buffer[s] += o->noise * (o->a.value >> 10) >> 6;
		cs_RunRamp(&o->a, 1);
	}
}


static inline int cso_Inter(int16_t *d, unsigned ph)
{
#ifdef CS_HIFI		/* Hermite */
	int i = ph >> 8, x = (ph & 0xff) << 7;
	int c = (d[i + 1] - d[i - 1]) >> 1;
	int a = (3 * (d[i] - d[i + 1]) + d[i + 2] - d[i - 1]) >> 1;
	int b = d[i - 1] - d[i] + c - a;
	a = a * x >> 15;
	a = (a + b) * x >> 15;
	return d[i] + ((a + c) * x >> 15);
#else			/* Linear */
	int i = ph >> 8, x = ph & 0xff;
	return (d[i] * (256 - x) + d[i + 1] * x) >> 8;
#endif
}


void cso_Wavetable(CS_oscillator *o, int *buffer, unsigned frames)
{
	CS_wave *w = o->wave;
	unsigned dph = o->dphase * w->period >> 8, mm, s, ph;
	int16_t *d;
	cs_PrepareRamp(&o->a, frames);
	/* FIXME: Cache, or do something smarter... */
	for(mm = 0; (dph > CS_MAXPHINC) && (mm < CS_MIPLEVELS - 1); ++mm)
		dph >>= 1;
	ph = o->phase >> mm;
	if(w->looped)
		ph %= w->size[mm] << 8;
	else if((ph >> 8) > (w->size[mm] + CS_WAVEPRE))
		return;		/* All played! */
	if(dph > CS_MAXPHINC)
	{
		ph += dph * frames;
		o->phase = ph << mm;
		cs_RunRamp(&o->a, frames);
		return;
	}
	d = w->data[mm] + CS_WAVEPRE;
	for(s = 0; s < frames; ++s)
	{
		int out = cso_Inter(d, ph) + cso_Inter(d, ph + (dph >> 1));
		buffer[s] += (int64_t)out * o->a.value >> (16 + 1);
		ph += dph;
		cs_RunRamp(&o->a, 1);
	}
	o->phase = ph << mm;
}


static inline void cs_OscReset(CS_oscillator *o)
{
	o->phase = 0;
	o->noise = 0;
	o->wave = NULL;
	o->Process = cso_Off;
	cs_SetRamp(&o->a, 0, 0);
}


static inline void cs_OscFrequency(CS_state *st, CS_oscillator *o, float f)
{
	o->dphase = f * 65536.0f / st->audiospec.freq;
}


/*
 *	ph	desired phase, [0, 1], (16):16 fixp
 *	sst	SubSample Time, [0, 1], (24):8 fixp
 */
static inline void cs_OscPhase(CS_oscillator *o, int ph, unsigned sst)
{
	if(!o->wave)
	{
		o->phase = 0;
		return;
	}
	o->phase = ph * o->wave->period >> 8;
	o->phase -= (sst * o->dphase) * o->wave->period >> 16;
}


/*---------------------------------------------------------
	Pan/volume output stage
---------------------------------------------------------*/

static void cs_PanmixProcess(CS_panmixer *pmx, int *in, int *out, unsigned frames)
{
	unsigned s;
	cs_PrepareRamp(&pmx->vol, frames);
	for(s = 0; s < frames; ++s)
	{
		out[s] += (int64_t)in[s] * pmx->vol.value >> 24;
		cs_RunRamp(&pmx->vol, 1);
	}
}


static inline void cs_PanmixReset(CS_panmixer *pmx)
{
	cs_SetRamp(&pmx->vol, 65536, 0);
	cs_SetRamp(&pmx->pan, 0, 0);
}


/*---------------------------------------------------------
	Voice and audio processing
---------------------------------------------------------*/

static inline CS_errors cs_VoicePush(CS_state *st, CS_voice *v, int firstreg,
		int interrupt)
{
	CS_stackentry *se = &cs_AllocBlock(st)->stackentry;
	if(!se)
		return CS_MEMORY;
	se->prev = v->stack;
	v->stack = se;
	se->program = v->program;
	se->state = v->state;
	se->pc = v->pc;
	if((se->interrupt = interrupt))
		se->timer = v->timer;
	se->firstreg = firstreg;
	memcpy(se->r, v->r + firstreg, sizeof(int) * (CS_REGISTERS - firstreg));
	return CS_OK;
}


static inline void cs_VoicePop(CS_state *st, CS_voice *v)
{
	CS_stackentry *se = v->stack;
	v->program = se->program;
	v->state = se->state;
	v->pc = se->pc;
	if(se->interrupt)
		v->timer = se->timer;
	memcpy(v->r + se->firstreg, se->r,
			sizeof(int) * (CS_REGISTERS - se->firstreg));
	v->stack = se->prev;
	cs_FreeBlock(st, se);
}


static void cs_VoiceControl(CS_state *st, CS_voice *v, unsigned reg, int frames)
{
	switch(reg)
	{
	  case R_TICK:
/*TODO: Adjust pending TDELAYs in subvoices...? */
		break;
	  case R_WAVE:
		cs_OscWave(&v->osc, st, v->r[reg] >> 16);
		break;
	  case R_PITCH:
	  case R_SPITCH:
		cs_OscFrequency(st, &v->osc, powf(2.0f,
				(v->r[R_PITCH] + v->r[R_SPITCH]) *
				(1.0f / 65536.0f)) * CS_MIDDLEC);
		break;
	  case R_AMPLITUDE:
	  case R_SAMPLITUDE:
		cs_SetRamp(&v->osc.a, (int64_t)v->r[R_AMPLITUDE] *
				v->r[R_SAMPLITUDE] >> 16,
				frames);
		break;
	  case R_VOLUME:
		cs_SetRamp(&v->pmx.vol, v->r[reg], frames);
		break;
	  case R_PANX:
		cs_SetRamp(&v->pmx.pan, v->r[reg], frames);
		break;
	}
}


static inline unsigned cs_VoiceTicks2t(CS_state *st, CS_voice *v, int d)
{
	return ((uint64_t)d * v->r[R_TICK] >> 8) * st->msdur >> 32;
}


/* WARNING: These are tuned for minimal init/cleanup overhead! Be careful... */
static CS_voice *cs_VoiceAlloc(CS_state *st)
{
	CS_voice *v = (CS_voice *)malloc(sizeof(CS_voice));
	if(!v)
		return NULL;
	v->sub = NULL;
	v->stack = NULL;
	v->program = NULL;
	v->messages = NULL;
	memset(v->r, 0, sizeof(int) * CS_CREGISTERS);
	v->r[R_VOLUME] = 65536;
	memset(v->sv, 0, sizeof(v->sv));
	cs_OscReset(&v->osc);
	cs_PanmixReset(&v->pmx);
	v->insert_cb = NULL;
	++st->totalvoices;
#ifdef DEBUG
	if(st->running)
		fprintf(stderr, "ChipSound: Voice pool exhausted! "
				"Allocated new voice %p.\n", v);
#endif
	return v;
}


CS_voice *cs_VoiceNew(CS_state *st, CS_voice *parent)
{
	CS_voice *v = st->voicepool;
	if(v)
		st->voicepool = v->next;
	else if(!(v = cs_VoiceAlloc(st)))
		return NULL;
	++st->activevoices;
	v->timer = 0;
	if(parent)
	{	/* Always, except "root" init, where everything's zero/NULL! */
		v->next = parent->sub;
		parent->sub = v;
		v->r[R_TICK] = parent->r[R_TICK];
		v->r[R_SPITCH] = parent->r[R_SPITCH];
		v->r[R_SAMPLITUDE] = parent->r[R_SAMPLITUDE];
	}
	cs_OscFrequency(st, &v->osc, powf(2.0f,
			v->r[R_SPITCH] * (1.0f / 65536.0f)) * CS_MIDDLEC);
	return v;
}


void cs_VoiceFree(CS_state *st, CS_voice **head)
{
	CS_voice *v = *head;
	cs_VoiceKill(st, v);
	*head = v->next;
	v->next = st->voicepool;
	st->voicepool = v;
	--st->activevoices;
	v->program = NULL;
	v->r[R_WAVE] = v->r[R_PITCH] = v->r[R_AMPLITUDE] = v->r[R_PANX] = 0;
	v->r[R_VOLUME] = 65536;
	cs_OscReset(&v->osc);
	cs_PanmixReset(&v->pmx);
	v->insert_cb = NULL;
	while(v->messages)
	{
		CS_message *m = v->messages;
		v->messages = m->next;
		cs_FreeMessage(st, m);
	}
#ifdef DEBUG
	{
		int i;
		assert(!v->sub);
		assert(!v->stack);
		for(i = 0; i < CS_REGISTERS; ++i)
			assert(!v->sv[i]);
	}
#endif
}


void cs_VoiceKill(CS_state *st, CS_voice *v)
{
	while(v->stack)
		cs_VoicePop(st, v);
	v->program = st->terminator;
	v->pc = v->program->eps[0].position;
	v->timer = 0;
	v->state = CS_RUNNING;
	while(v->sub)
		cs_VoiceFree(st, &v->sub);
	memset(v->sv, 0, sizeof(v->sv));
}
/* /WARNING */


CS_errors cs_VoiceCall(CS_state *st, CS_voice *v, CS_program *p, unsigned ep,
		int argc, int *argv)
{
	int i;
	if(p->eps[ep].position == 0xffffffff)
		return CS_BADENTRY;
	if(v->program)
		if(cs_VoicePush(st, v, p->eps[ep].argv, (ep > 0)))
			return CS_MEMORY;
	v->program = p;
	v->pc = p->eps[ep].position;
	v->state = ep > 0 ? CS_INTERRUPT : CS_RUNNING;
	if(argc > p->eps[ep].argc)
		argc = p->eps[ep].argc;
	memcpy(v->r + p->eps[ep].argv, argv, argc * sizeof(int));
	for(i = argc; i < p->eps[ep].argc; ++i)
		v->r[i + p->eps[ep].argv] = p->eps[ep].argdefs[i];
//printf("%p: VoiceCall: %d\n", v, v->state);
	return CS_OK;
}


static inline CS_errors cs_VoiceSend(CS_state *st, CS_voice *v, unsigned when,
		unsigned ep, int argc, int *argv)
{
	int res;
	CS_message *m;
	if(!(m = cs_AllocMessage(st)))
		return CS_MEMORY;
	m->m.timestamp = when + st->now_fragstart;
	m->m.what = CSM_MESSAGE;
	m->m.addr = ep;
	m->m.argc = argc;
	memcpy(m->m.a, argv, argc * sizeof(int));
	cs_SendMessage(v, m);
	return CS_OK;
}


static CS_errors cs_VoiceSpawn(CS_state *st, CS_voice *v, unsigned when,
		int vid, unsigned program, int argc, int *argv)
{
	int res;
	CS_voice *nv;
	CS_program *p = (CS_program *)cs_ObjGet(st, program, CS_TPROGRAM);
	if((vid > 0) && (v->sv[vid]))	/* NOTE: No detach for id 0! */
		cs_VoiceDetach(v->sv[vid]);
	if(!p)
		return CS_BADPROGRAM;
	nv = cs_VoiceNew(st, v);
	if(vid > 0)
		v->sv[vid] = nv;
	if(!nv)
		return CS_VOICEALLOC;
	nv->id = vid;
	DBG(assert(nv->id);)
	nv->timer = when;
	if((res = cs_VoiceCall(st, nv, p, 0, argc, argv)))
		cs_VoiceFree(st, &v->sub);
	return res;
}


static inline CS_errors csr_Starta(CS_state *st, CS_messagebody *m)
{
	CS_voice *v;
	CS_program *p = (CS_program *)cs_ObjGet(st, m->addr, CS_TPROGRAM);
	if(!p)
		return CS_BADPROGRAM;
	if(!(v = cs_VoiceNew(st, st->groups[m->gid])))
		return CS_VOICEALLOC;
	v->id = m->vid ? m->vid : -1;
	v->timer = m->timestamp - st->now_fragstart;
//printf(" === starting voice %p, id %d ===\n", v, v->id);
if(v->timer > 0x7fffffff)
{
#if TIMESTAMPWARNINGS == 1
	printf("ChipSound: Incorrect timestamp in csr_Starta()! (%f frames late.)\n",
			(st->now_fragstart - m->timestamp) / 256.0f);
#endif
	v->timer = 0;
}
	return cs_VoiceCall(st, v, p, 0, m->argc, m->a);
}


DUMPMSGS(static void printargs(int argc, int *argv)
{
	int i;
	for(i = 0; i < argc; ++i)
		if(i < argc - 1)
			printf("%f, ", argv[i] / 65536.0f);
		else
			printf("%f", argv[i] / 65536.0f);
})

/*
 * Process messages. Will return when:
 *	* VM needs to run to respond to message. (Returns CS_OK)
 *	* Time/audio needs to advance. (Returns CS_END)
 *	* There are no more messages in the queue. (Returns CS_END)
 *	* There is an error. (Returns an error code)
 *
TODO: Forward messages where applicable, to save a realloc + copy.
 */
static CS_errors cs_VoiceProcessMessages(CS_state *st, CS_voice *v)
{
	unsigned current = v->messages->m.timestamp >> 8;
	while(v->messages)
	{
		int res;
		CS_message *m = v->messages;
		if((m->m.timestamp >> 8) != current)
			return CS_END;
		DUMPMSGS(printf("%f:\tGID %u VID %u (%p) ",
				m->m.timestamp / 256.0f, m->m.gid, m->m.vid, v);)
		NUMMSGS(printf("[ %u ] ", m->number);)
		switch(m->m.what)
		{
		  /* Sub-voice messages */
		  case CSM_STARTSUB:
			DUMPMSGS(printf("STARTSUB(");
					printargs(m->m.argc, m->m.a);
					printf(")\n");)
#if 0
			if((res = csr_Starta(st, &m->m)))
				return res;
#else
			csr_Starta(st, &m->m);
#endif
			break;
		  case CSM_SENDSUB:
		  {
			CS_voice *sv;
			unsigned when = m->m.timestamp - st->now_fragstart;
if(when > 0x7fffffff)
{
#if TIMESTAMPWARNINGS == 1
	printf("ChipSound: Incorrect CSM_SENDSUB timestamp! (%f frames late.)\n",
			(st->now_fragstart - m->m.timestamp) / 256.0f);
#endif
	when = 0;
}
			DUMPMSGS(printf("SENDSUB(%u: ", m->m.addr);
					printargs(m->m.argc, m->m.a);
					printf(") to");)
			for(sv = v->sub; sv; sv = sv->next)
				if(!m->m.vid || sv->id == m->m.vid)
				{
					DUMPMSGS(printf("\t%d (%p)", sv->id, sv);)
					cs_VoiceSend(st, sv, when, m->m.addr,
							m->m.argc, m->m.a);
				}
				DUMPMSGS(printf("\n");)
			break;
		  }
		  case CSM_DETACHSUB:
		  {
			CS_voice *sv;
			DUMPMSGS(printf("DETACHSUB to");)
			for(sv = v->sub; sv; sv = sv->next)
			{
				CS_message *msg;
				DUMPMSGS(printf("\t%d (%p)", sv->id, sv);)
				if(m->m.vid && (sv->id != m->m.vid))
					continue;
				if(!(msg = cs_AllocMessage(st)))
					return CS_MEMORY;
				msg->m.timestamp = m->m.timestamp;
if(cs_TSDiff(msg->m.timestamp, st->now_fragstart) < 0)
{
#if TIMESTAMPWARNINGS == 1
	printf("ChipSound: Incorrect CSM_DETACHSUB timestamp! (%f frames late.)\n",
			(st->now_fragstart - msg->m.timestamp) / 256.0f);
#endif
	msg->m.timestamp = st->now_fragstart;
}
				msg->m.what = CSM_DETACH;
				cs_SendMessage(sv, msg);
			}
			DUMPMSGS(printf("\n");)
			break;
		  }

		  /* Direct voice messages */
		  case CSM_MESSAGE:
			DUMPMSGS(printf("MESSAGE(%u: ", m->m.addr);
					printargs(m->m.argc, m->m.a);
					printf(")\n");)
			res = cs_VoiceCall(st, v, v->program, m->m.addr,
					m->m.argc, m->m.a);
			if(res == CS_BADENTRY)
				break;	/* No entry point - ignore! */
			v->timer = m->m.timestamp & 0xff;
if(cs_TSDiff(m->m.timestamp, st->now_fragstart) < 0)
{
#if TIMESTAMPWARNINGS == 1
	printf("ChipSound: Incorrect CSM_MESSAGE timestamp! (%f frames late.)\n",
			(st->now_fragstart - m->m.timestamp) / 256.0f);
#endif
	v->timer = 0;
}
			v->messages = m->next;
			cs_FreeMessage(st, m);
			return res;
		  case CSM_DETACH:
			DUMPMSGS(printf("DETACH\n");)
			cs_VoiceDetach(v);
			v->messages = m->next;
			cs_FreeMessage(st, m);
			return CS_OK;
		}
		v->messages = m->next;
		cs_FreeMessage(st, m);
	}
	return CS_END;
}


/*
 * Execute VM instructions until a timing instruction is executed, or the
 * program ends. Returns CS_OK as long as the VM program wants to keep running.
 */
static inline CS_errors cs_VoiceVMProcess(CS_state *st, CS_voice *v, unsigned frame)
{
	int res, cargc = 0, cargv[CS_MAXARGS];
	unsigned *code = v->program->bank->code;
	int *r = v->r;
	unsigned inscount = CS_INSLIMIT;
	unsigned cregmask = 0;
	if(v->timer >= 256)
		return CS_OK;
	if(v->state == CS_WAITING)
		v->state = CS_RUNNING;
	while(1)
	{
		unsigned ins = code[v->pc];
		CS_opcodes op = ins >> 26;
		unsigned reg = (ins >> 21) & 0x1f;
		int arg = ins & 0x1fffff;
		DUMPCODE(printf("%p: ", v); cs_DumpIns(code, v->pc);)
		if(!--inscount)
		{
			cs_VoiceKill(st, v);
			st->instructions += CS_INSLIMIT;
			return CS_OVERLOAD;
		}
		switch(op)
		{
		  case OP_END:
			if(v->stack)
			{
				cs_VoicePop(st, v);
				code = v->program->bank->code;
				switch(v->state)
				{
				  case CS_WAITING:
					if(v->timer >= 256)
					{
						st->instructions +=
								(CS_INSLIMIT -
								inscount);
						return CS_OK;
					}
					v->state = CS_RUNNING;
					continue;
				  case CS_INTERRUPT:
				  case CS_RUNNING:
					continue;
				  case CS_ENDING:
				  case CS_FINALIZING:
					break;
				}
			}
			if(cregmask)
				for(reg = 0; reg < CS_CREGISTERS; ++reg)
					if(cregmask & (1 << reg))
						cs_VoiceControl(st, v, reg, 0);
			v->timer = 0xffffffff;
			if(v->state == CS_FINALIZING)
			{
				/* Wait for subvoices to terminate */
//printf("%p: --- END 1 - Wait for subvoices to terminate\n", v);
				st->instructions += (CS_INSLIMIT - inscount);
				return v->sub ? CS_OK : CS_END;
			}
			v->state = CS_ENDING;
			if((v->id != -1) || v->messages)
			{
				/* Hang around until detached! */
//printf("%p: --- END 2 - Hang around until detached\n", v);
				st->instructions += (CS_INSLIMIT - inscount);
				return CS_OK;
			}
			v->state = CS_FINALIZING;
			if(!v->sub)
			{
				/* That's it - all done! */
//printf("%p: --- END 3 - That's it - all done!\n", v);
				st->instructions += (CS_INSLIMIT - inscount);
				return CS_END;
			}
			/* Detach subvoices, then wait for them to terminate */
//printf("%p: --- END 4 - Detach subvoices, then wait for them to terminate\n", v);
			memset(v->sv, 0, sizeof(v->sv));
			for(v = v->sub; v; v = v->next)
				cs_VoiceDetach(v);
			st->instructions += (CS_INSLIMIT - inscount);
			return CS_OK;
		  case OP_JUMP:
			v->pc = arg;
			continue;
		  case OP_LOOP:
			r[reg] -= 65536;
			if(r[reg] <= 0)
				break;
			v->pc = arg;
			continue;
		  case OP_JZ:
			if(r[reg])
				break;
			v->pc = arg;
			continue;
		  case OP_JNZ:
			if(!r[reg])
				break;
			v->pc = arg;
			continue;
		  case OP_JG:
			if(r[reg] <= 0)
				break;
			v->pc = arg;
			continue;
		  case OP_JL:
			if(r[reg] >= 0)
				break;
			v->pc = arg;
			continue;
		  case OP_JGE:
			if(r[reg] < 0)
				break;
			v->pc = arg;
			continue;
		  case OP_JLE:
			if(r[reg] > 0)
				break;
			v->pc = arg;
			continue;
		  case OP_WAKE:
		  {
			CS_stackentry *se = v->stack;
			while(se->prev && (se->state == CS_INTERRUPT))
				se = se->prev;
			if(se->state < CS_ENDING)
				break;
			/* Fall through for the actual wakeup! */
		  }
		  case OP_FORCE:
		  {
			CS_stackentry *se = v->stack;
			while(se->prev && (se->state == CS_INTERRUPT))
				se = se->prev;
			se->pc = arg;
			se->timer = 0;
			se->state = CS_RUNNING;
			v->r[R_AMPLITUDE] = v->osc.a.value >> 8;
			v->r[R_VOLUME] = v->pmx.vol.value >> 8;
			v->r[R_PANX] = v->pmx.pan.value >> 8;
			break;
		  }
		  case OP_SUBR:
			r[reg] -= r[arg];
			cregmask |= 1 << reg;
			break;
		  case OP_DIVR:
			if(!r[arg])
			{
				st->instructions += (CS_INSLIMIT - inscount);
				return CS_DIVBYZERO;
			}
			r[reg] = ((int64_t)r[reg] << 16) / r[arg];
			cregmask |= 1 << reg;
			break;
		  case OP_P2DR:
			r[reg] = 65536000.0f / (powf(2.0f, r[arg] *
					(1.0f / 65536.0f)) * CS_MIDDLEC);
			cregmask |= 1 << reg;
			break;
		  case OP_DELAY:
			v->timer += (int64_t)cs_f2i(arg) * st->msdur >> 24;
			goto timing;
		  case OP_DELAYR:
			v->timer += (int64_t)r[reg] * st->msdur >> 24;
			goto timing;
		  case OP_TDELAY:
			v->timer += cs_VoiceTicks2t(st, v, cs_f2i(arg));
			goto timing;
		  case OP_TDELAYR:
			v->timer += cs_VoiceTicks2t(st, v, r[reg]);
			goto timing;
		  case OP_PHASE:
			cs_OscPhase(&v->osc, cs_f2i(arg), v->timer);
			break;
		  case OP_PHASER:
			cs_OscPhase(&v->osc, r[reg], v->timer);
			break;
		  case OP_DEBUGR:
			printf(":: ChipSound DEBUG: R%d=%f\t(%p)\n", reg,
					r[reg] * (1.0f / 65536.0f), v);
			break;
		  case OP_DEBUG:
			printf(":: ChipSound DEBUG: %f\t(%p)\n",
					cs_f2i(arg) * (1.0f / 65536.0f), v);
			break;
		  case OP_LOAD:
			r[reg] = cs_f2i(arg);
			cregmask |= 1 << reg;
			break;
		  case OP_LOADR:
			r[reg] = r[arg];
			cregmask |= 1 << reg;
			break;
		  case OP_ADD:
			r[reg] += cs_f2i(arg);
			cregmask |= 1 << reg;
			break;
		  case OP_ADDR:
			r[reg] += r[arg];
			cregmask |= 1 << reg;
			break;
		  case OP_MUL:
			r[reg] = (int64_t)r[reg] * cs_f2i(arg) >> 16;
			cregmask |= 1 << reg;
			break;
		  case OP_MULR:
			r[reg] = (int64_t)r[reg] * r[arg] >> 16;
			cregmask |= 1 << reg;
			break;
		  case OP_MOD:
			r[reg] %= cs_f2i(arg);
			cregmask |= 1 << reg;
			break;
		  case OP_MODR:
			if(!r[arg])
			{
				st->instructions += (CS_INSLIMIT - inscount);
				return CS_DIVBYZERO;
			}
			r[reg] %= r[arg];
			cregmask |= 1 << reg;
			break;
		  case OP_QUANT:
			arg = cs_f2i(arg);
			r[reg] = r[reg] / arg * arg;
			cregmask |= 1 << reg;
			break;
		  case OP_QUANTR:
			r[reg] = r[reg] / r[arg] * r[arg];
			cregmask |= 1 << reg;
			break;
		  case OP_RAND:
			r[reg] = (int64_t)cs_Noise(st) * cs_f2i(arg) >> 16;
			cregmask |= 1 << reg;
			break;
		  case OP_RANDR:
			r[reg] = (int64_t)cs_Noise(st) * r[arg] >> 16;
			cregmask |= 1 << reg;
			break;
		  case OP_PUSH:
			if(cargc >= CS_MAXARGS)
			{
				st->instructions += (CS_INSLIMIT - inscount);
				return CS_MANYARGS;
			}
			cargv[cargc++] = cs_f2i(arg);
			break;
		  case OP_PUSHR:
			if(cargc >= CS_MAXARGS)
			{
				st->instructions += (CS_INSLIMIT - inscount);
				return CS_MANYARGS;
			}
			cargv[cargc++] = r[reg];
			break;
		  case OP_RUNR:
			arg = r[arg] >> 16; /* Odd encoding to match SPAWN(R) */
		  case OP_RUN:
		  {
			CS_program *p = (CS_program *)cs_ObjGet(st, arg,
					CS_TPROGRAM);
			if(!p)
			{
				st->instructions += (CS_INSLIMIT - inscount);
				return CS_BADPROGRAM;
			}
			++v->pc;	/* Return to *next* instruction! */
			if((res = cs_VoiceCall(st, v, p, 0, cargc, cargv)))
			{
				st->instructions += (CS_INSLIMIT - inscount);
				return res;
			}
			cargc = 0;
			code = v->program->bank->code;
			continue;
		  }
		  case OP_SPAWNR:
			arg = r[arg] >> 16;
		  case OP_SPAWN:
			cs_VoiceSpawn(st, v, (frame << 8) + v->timer,
					reg, arg, cargc, cargv);
			cargc = 0;
			break;
		  case OP_SPAWNDR:
			arg = r[arg] >> 16;
		  case OP_SPAWND:
			cs_VoiceSpawn(st, v, (frame << 8) + v->timer,
					-1, arg, cargc, cargv);
			cargc = 0;
			break;
#if 0
		  case OP_SENDR:
			reg = r[reg] >> 16;
			if(reg > CS_REGISTERS)
				return CS_BADVOICE;
#endif
		  case OP_SEND:
			DBG(if(!arg)
				printf("Weird...! OP_SEND to EP0...\n");)
			if(v->sv[reg])
				cs_VoiceSend(st, v->sv[reg],
						(frame << 8) + v->timer,
						arg, cargc, cargv);
//printf("%p: SEND msg %d to %p at %f\n", v, arg, v->sv[reg], ((frame << 8) + v->timer) / 256.0f);
			cargc = 0;
			break;
		  case OP_SENDA:
		  {
			CS_voice *sv;
			unsigned when = (frame << 8) + v->timer;
			for(sv = v->sub; sv; sv = sv->next)
				cs_VoiceSend(st, sv, when, arg, cargc, cargv);
			cargc = 0;
			break;
		  }
		  case OP_SENDS:
			if((res = cs_VoiceCall(st, v, v->program, arg, cargc,
					cargv)))
			{
				st->instructions += (CS_INSLIMIT - inscount);
				return res;
			}
			cargc = 0;
			break;
		  case OP_SET:
			if(reg >= CS_CREGISTERS)
				break;
			cs_VoiceControl(st, v, reg, 0);
			cregmask &= ~(1 << reg);
			break;
#if 0
		  case OP_RAMP:
			if(reg >= CS_CREGISTERS)
				break;
			arg = (int64_t)cs_f2i(arg) * st->msdur >> 32;
			cs_VoiceControl(v, reg, arg);
			cregmask &= ~(1 << reg);
			break;
		  case OP_RAMPR:
			if(reg >= CS_CREGISTERS)
				break;
			arg = (int64_t)r[arg] * st->msdur >> 32;
			cs_VoiceControl(v, reg, arg);
			cregmask &= ~(1 << reg);
			break;
		  case OP_DETACHR:
			arg = r[arg] >> 16;
		  case OP_DETACH:
			if(v->sv[arg])
				cs_VoiceDetach(v->sv[arg]);
			v->sv[arg] = NULL;
			break;
#endif
		  case OP_KILL:
			if(v->sv[reg])
				cs_VoiceKill(st, v->sv[reg]);
			v->sv[reg] = NULL;
			break;
		  case OP_KILLA:
		  {
			CS_voice *sv;
			for(sv = v->sub; sv; sv = sv->next)
				cs_VoiceKill(st, sv);
			memset(v->sv, 0, sizeof(v->sv));
			break;
		  }
		}
		++v->pc;
		continue;
	  timing:
		++v->pc;
		if(v->timer >= 256)
		{
			int d = v->timer >> 8;
			for(reg = 0; reg < CS_CREGISTERS; ++reg)
				if(cregmask & (1 << reg))
					cs_VoiceControl(st, v, reg, d);
			v->state = CS_WAITING;
			st->instructions += (CS_INSLIMIT - inscount);
			return CS_OK;
		}
		continue;
	}
}


static inline CS_errors cs_VoiceProcess(CS_state *st, CS_voice *v,
		int *buf, unsigned frames)
{
	int s_end = 0;
	int s = 0;
	while(s < frames)
	{
		int vmres;
		while(1)
		{
			int res, mt;
			if((vmres = cs_VoiceVMProcess(st, v, s)) > CS_END)
				return vmres;
			s_end = s + (v->timer >> 8);
			if(!v->messages)
				break;
			mt = cs_TSDiff(v->messages->m.timestamp,
					st->now_fragstart + (s << 8));
			if(mt >= 256)
			{
				if(mt < v->timer)
					s_end = s + (mt >> 8);
				break;
			}
			if((res = cs_VoiceProcessMessages(st, v)))
			{
				DBG(if(res > CS_END)
					printf("cs_VoiceProcessMessages(): %s!\n",
							cs_ErrorString(res));)
				break;
			}
		}
		if(vmres == CS_END)
			return CS_END;
		if(s_end > frames)
			s_end = frames;
		v->osc.Process(&v->osc, buf + s, s_end - s);
		v->timer -= (s_end - s) << 8;
		s = s_end;
	}
	return CS_OK;
}


static void cs_ProcessVoices(CS_state *st, CS_voice **head,
		int *buf, unsigned frames);


static inline void cs_ProcessSubvoices(CS_state *st, CS_voice *v,
		int *buf, unsigned frames)
{
	if(!v->sub)
		return;
	cs_ProcessVoices(st, &v->sub, buf, frames);
	if(!v->sub)
		if(v->state >= CS_ENDING)
			v->timer = 0;	/* Notify parent that subs are done! */
}


static void cs_ProcessVoices(CS_state *st, CS_voice **head,
		int *buf, unsigned frames)
{
	while(*head)
	{
		CS_voice *v = *head;
		CS_errors res;
		int *workbuf;
/*TODO: Keep allocated buffers until we're done with all voices! Just add more
 *TODO: as needed, and free them before we return. */
		if(CS_NEEDS_WORKBUFS(v))
		{
			workbuf = cs_AllocBlock(st)->buffer;
			memset(workbuf, 0, frames * sizeof(int));
		}
		else
			workbuf = buf;
		if((res = cs_VoiceProcess(st, v, workbuf, frames)))
		{
			if(res != CS_END)
				fprintf(stderr, "ChipSound: VM error %s!\n",
						cs_ErrorString(res));
			cs_VoiceFree(st, head);
		}
		else
		{
			cs_ProcessSubvoices(st, v, workbuf, frames);
			if(v->insert_cb)
				v->insert_cb(&workbuf, 1, frames, v->insert_ud);
			if(workbuf != buf)
				cs_PanmixProcess(&v->pmx, workbuf, buf, frames);
			head = &v->next;
		}
		if(workbuf != buf)
			cs_FreeBlock(st, workbuf);
	}
}


static void cs_ProcessMaster(CS_state *st, int *in, Sint16 *out, unsigned frames)
{
	int i;
	/* mono -> stereo delay + 6:26 --> 16:16 conversion */
	for(i = 0; i < frames; ++i)
	{
		int outl, outr;
		outl = outr = in[i];
		/* Stereo feedback delay */
#if 1
#define	_DB(x)	st->delaybuf[(st->delaypos - (x)) & 0xffff]
		_DB(0) = (outl + _DB(st->fbdelay) +
				_DB(st->fbdelay / 3)) >> 2;
		outl += _DB(st->ldelay) >> 2;
		outl += (_DB(st->ldelay / 3) +
				_DB(st->ldelay / 5) +
				_DB(st->ldelay / 8)) >> 1;
		outr += _DB(st->rdelay) >> 2;
		outr += (_DB(st->rdelay / 4) +
				_DB(st->rdelay / 6) +
				_DB(st->rdelay / 7)) >> 1;
		++st->delaypos;
#endif
		/* Scale to 16:16 */
		st->masterbuf[i * 2] = outl >> 10;
		st->masterbuf[i * 2 + 1] = outr >> 10;
	}

	/* Stereo effect processing */
	limss_process(&st->limiter, st->masterbuf, st->masterbuf, frames);

	/* Clipping + 16 bit output conversion */
	for(i = 0; i < frames; ++i)
	{
		int outl = st->masterbuf[i * 2];
		int outr = st->masterbuf[i * 2 + 1];

		/* Clipping */
		if(outl < -32768)
			outl = 32768;
		else if(outl > 32767)
			outl = 32767;
		if(outr < -32768)
			outr = 32768;
		else if(outr > 32767)
			outr = 32767;

		/* Output */
		out[i * 2] = outl;
		out[i * 2 + 1] = outr;
	}
}


static void cs_ProcessMIDI(CS_state *st)
{
#ifdef CS_ALSA_MIDI
	unsigned ch;
	snd_seq_event_t *ev;
	if(!st->midiseqh)
		return;
	if(snd_seq_event_input(st->midiseqh, &ev) < 0)
		return;
	ch = ev->data.control.channel;
	if(ch >= 16)
	{
		snd_seq_free_event(ev);
		return;
	}
	switch(ev->type)
	{
	  case SND_SEQ_EVENT_CONTROLLER:
	  {
		switch(ev->data.control.param)
		{
		  case 1:
			st->midimod[ch] = ev->data.control.value;
			cs_Send(MIDIGRP, 0, 3, st->midimod[ch] / 127.0f);
			break;
		}
		break;
	  }
	  case SND_SEQ_EVENT_PGMCHANGE:
		st->midiprg[ch] = ev->data.control.value;
#ifdef DEBUG
		printf("MIDI: PgmChange %d:'%s' (%s).\n", st->midiprg[ch],
				cs_GetName(-1, st->midiprg[ch]),
				cs_TypeName(cs_GetType(cs_Get(-1,
				st->midiprg[ch]))));
#endif
		break;
	  case SND_SEQ_EVENT_NOTEON:
		if(ev->data.note.velocity)
		{
			cs_Start(MIDIGRP, (ch << 7) | ev->data.note.note,
					cs_Get(-1, st->midiprg[ch]),
				 	(ev->data.note.note - 60) / 12.0f,
					ev->data.note.velocity / 127.0f,
					st->midimod[ch] / 127.0f);
			break;
		}
		/* Fall through for NoteOn(pitch,0)! */
	  case SND_SEQ_EVENT_NOTEOFF:
		cs_Send(MIDIGRP, (ch << 7) | ev->data.note.note, 1);
		cs_Detach(MIDIGRP, (ch << 7) | ev->data.note.note);
		break;
/*	  case SND_SEQ_EVENT_KEYPRESS:
		ev_aftertouch(d->msg, ev->data.control.channel,
				ev->data.note.note,
				ev->data.note.velocity);
		break;
	  case SND_SEQ_EVENT_CHANPRESS:
		ev_pressure(d->msg, ev->data.control.channel,
				ev->data.control.value);
		break;
	  case SND_SEQ_EVENT_PITCHBEND:
		ev_bend(d->msg, ev->data.control.channel,
				ev->data.control.value);
		break;*/
	}
	snd_seq_free_event(ev);
#endif
}


static void cs_AudioCallback(void *ud, Uint8 *stream, int len)
{
	CS_state *st = &cs_globalstate;
	Sint16 *devicebuf = (Sint16 *)stream;
	int remain = len / 4;
	unsigned t1 = SDL_GetTicks();
	unsigned dur;
	int *mixbuf = cs_AllocBlock(st)->buffer;
	if(!mixbuf)
		return;
	cs_PumpMessages(st);
	cs_ProcessMIDI(st);
	while(remain)
	{
		unsigned frag = remain > CS_MAXFRAG ? CS_MAXFRAG : remain;
		memset(mixbuf, 0, frag * sizeof(int));
		cs_ProcessVoices(st, &st->groups[0], mixbuf, frag);
		cs_ProcessMaster(st, mixbuf, devicebuf, frag);
		devicebuf += frag * 2;
		remain -= frag;
		st->now_fragstart += frag << 8;
	}
	cs_FreeBlock(st, mixbuf);
	dur = SDL_GetTicks() - t1;
	if(dur > st->cputimemax)
		st->cputimemax = dur;
	st->cputimesum += dur;
	++st->cputimecount;
	if(dur)
	{
		unsigned ld = (t1 - st->now_ticks) * 100 / dur;
		if(ld > st->cpuloadmax)
			st->cpuloadmax = ld;
	}
	st->now_frames = st->now_fragstart;
	st->now_ticks = t1;
	st->now_guard = st->now_frames;
}


/*---------------------------------------------------------
	Versioning
---------------------------------------------------------*/

unsigned cs_Version(void)
{
	return CS_VERSION;
}


/*---------------------------------------------------------
	Open/close
---------------------------------------------------------*/

static CS_errors cs_InitAudio(CS_state *st, int samplerate, int buffer,
		int channels)
{
	SDL_AudioSpec as;
	if(SDL_InitSubSystem(SDL_INIT_AUDIO) < 0)
		return CS_DEVICEOPEN;
	as.freq = samplerate;
	as.format = AUDIO_S16SYS;
	as.channels = channels;
	as.samples = buffer;
	as.callback = cs_AudioCallback;
	if(SDL_OpenAudio(&as, &st->audiospec) < 0)
	{
		fprintf(stderr, "SDL error: %s\n", SDL_GetError());
		return CS_DEVICEOPEN;
	}
	if(st->audiospec.format != AUDIO_S16SYS)
		return CS_BADFORMAT;
	st->delaybuf = calloc(65536, sizeof(int));
	st->masterbuf = calloc(CS_MAXFRAG * 2, sizeof(int));
	if(!st->delaybuf || !st->masterbuf)
		return CS_MEMORY;
	st->fbdelay = st->audiospec.freq * 0.4;
	st->rdelay = st->audiospec.freq * 0.28;
	st->ldelay = st->audiospec.freq * 0.32;
	st->rnd = 16576;
	lim_open(&st->limiter, st->audiospec.freq);
	lim_control(&st->limiter, LIM_THRESHOLD, 32768);
	lim_control(&st->limiter, LIM_RELEASE, 5000000);
	return CS_OK;
}


static CS_errors cs_InitMIDI(CS_state *st)
{
#ifdef CS_ALSA_MIDI
	int acaps = SND_SEQ_PORT_CAP_WRITE | SND_SEQ_PORT_CAP_SUBS_WRITE;
	const char *label = "ChipSound MIDI";
	if(snd_seq_open(&st->midiseqh, "default", SND_SEQ_OPEN_DUPLEX,
			SND_SEQ_NONBLOCK) < 0)
	{
		fprintf(stderr, "ChipSound: Couldn't open ALSA sequencer!\n");
		return CS_DEVICEOPEN;
	}
	snd_seq_set_client_name(st->midiseqh, label);
	st->midiportid = snd_seq_create_simple_port(st->midiseqh, label, acaps,
					SND_SEQ_PORT_TYPE_SYNTH);
	if(st->midiportid < 0)
	{
		fprintf(stderr, "ChipSound: Couldn't create MIDI port!'.\n");
		snd_seq_close(st->midiseqh);
		st->midiseqh = NULL;
		return CS_DEVICEOPEN;
	}
	cs_NewGroup(0, MIDIGRP);
#endif
	return CS_OK;
}


static CS_errors cs_InitWaves(void)
{
	int res = 0, i;
	int16_t buf[SC_WPER];
	res |= cs_Oscillator(0, "off", 256, cso_Off);
	for(i = 0; i < SC_WPER; ++i)
		buf[i] = i < SC_WPER / 2 ? 32767 : -32767;
	res |= cs_UploadWave(0, "square", SC_WPER, CS_LOOPED,
			CS_I16, buf, sizeof(buf));
	for(i = 0; i < SC_WPER; ++i)
		buf[i] = i * 65534 / SC_WPER - 32767;
	res |= cs_UploadWave(0, "saw", SC_WPER, CS_LOOPED,
			CS_I16, buf, sizeof(buf));
//	res |= cs_UploadWave(0, "sawimp", SC_WPER, 0, CS_I16, buf, sizeof(buf));
	for(i = 0; i < SC_WPER / 2; ++i)
		buf[(5 * SC_WPER / 4 - i - 1) % SC_WPER] =
				buf[i + SC_WPER / 4] =
				i * 65534 * 2 / SC_WPER - 32767;
	res |= cs_UploadWave(0, "triangle", SC_WPER, CS_LOOPED,
			CS_I16, buf, sizeof(buf));
	for(i = 0; i < SC_WPER; ++i)
		buf[i] = sin(i * 2.0f * M_PI / SC_WPER) * 32767.0f;
	res |= cs_UploadWave(0, "sine", SC_WPER, CS_LOOPED,
			CS_I16, buf, sizeof(buf));
	res |= cs_Oscillator(0, "noise", 256, cso_Noise);
#if 0
	for(i = 0; i < SC_WPER; ++i)
		buf[i] = sin(i * (1 + i * .1) * .0005) * 32767;
	res |= cs_UploadWave(0, "chirp", SC_WPER, CS_LOOPED,
			CS_I16, buf, sizeof(buf));
#endif
	return res < 0 ? CS_MEMORY : CS_OK;
}


static CS_errors cs_Open2(CS_state *st, int samplerate, int buffer,
		int channels, int flags)
{
	int i, res;
	if(st->running)
		cs_Close();
	memset(st, 0, sizeof(CS_state));
	st->flags = flags;
	for(i = 0; i < CS_INITBLOCKS; ++i)
	{
		CS_block *b = malloc(sizeof(CS_block));
		if(!b)
			return CS_MEMORY;
		b->next = st->blockpool;
		st->blockpool = b;
	}
	if((res = cs_InitAudio(st, samplerate, buffer, channels)) ||
			(res = -cs_NewBank("root")) || (res = cs_InitWaves()) ||
			(res = cs_OpenCompiler(st, flags)))
		return res;
	if((res = cs_CompileString(st->c, (CS_bank *)st->objects[0],
			"csgroupdriver()\n"
			"{\n"
			"	!V 1\n"
			".restart\n"
			"	vol V\n"
			"	d 100\n"
			"	1(NV) { V NV, force restart }\n"
			"}\n"
			"csterminator() { }\n", "builtin")))
		return res;
	if(!(st->groupprog = (CS_program *)cs_ObjGet(st, cs_Find(0,
			"csgroupdriver"), CS_TPROGRAM)))
		return CS_INTERNAL + 1;
	if(!(st->terminator = (CS_program *)cs_ObjGet(st, cs_Find(0,
			"csterminator"), CS_TPROGRAM)))
		return CS_INTERNAL + 2;
	for(i = 0; i < CS_INITVOICES; ++i)
	{
		CS_voice *v = cs_VoiceAlloc(st);
		if(!v)
			return CS_MEMORY;
		v->next = st->voicepool;
		st->voicepool = v;
	}
	st->msdur = st->audiospec.freq * 65.536f + .5f;
	if(!(st->groups[0] = cs_VoiceNew(st, NULL)))
		return CS_MEMORY;
	st->groups[0]->id = 0;
	st->groups[0]->next = NULL;
	st->groups[0]->r[R_TICK] = CS_DEFAULTTICK;
	for(i = 0; i < CS_CREGISTERS; ++i)
		cs_VoiceControl(st, st->groups[0], i, 0);
	if((res = cs_VoiceCall(st, st->groups[0], st->groupprog, 0, 0, NULL)))
		return res;
	if(flags & CS_MIDI)
		if((res = cs_InitMIDI(st)))
			return res;
	if((res = cs_OpenAPI(st, (float)buffer / samplerate)))
		return res;
	st->now_ticks = st->avgtick = SDL_GetTicks();
	SDL_PauseAudio(0);
	st->running = 1;
	return CS_OK;
}


CS_errors cs_Open(int samplerate, int buffer, int channels, int flags)
{
	int res = cs_Open2(&cs_globalstate, samplerate, buffer, channels, flags);
	if(res)
	{
		cs_Close();
		fprintf(stderr, "ChipSound: Initialization failed; %s!\n",
				cs_ErrorString(res));
		return res;
	}
	return CS_OK;
}


void cs_Close(void)
{
	CS_state *st = &cs_globalstate;
	int i;
	if(!st->running)
		return;
	SDL_CloseAudio();
	SDL_QuitSubSystem(SDL_INIT_AUDIO);
	cs_CloseAPI(st);
	free(st->masterbuf);
	free(st->delaybuf);
	lim_close(&st->limiter);
	if(st->groups[0])
		cs_VoiceFree(st, &st->groups[0]);
	while(st->voicepool)
	{
		CS_voice *v = st->voicepool;
		st->voicepool = v->next;
		free(v);
	}
	while(st->blockpool)
	{
		CS_block *b = st->blockpool;
		st->blockpool = b->next;
		free(b);
	}
	for(i = 0; i < st->maxobjects; ++i)
		cs_ObjFree(st, i);
	free(st->objects);
	while(st->banks)
	{
		CS_symbol *s = st->banks;
		st->banks = s->next;
		cs_FreeSymbol(s);
	}
	cs_CloseCompiler(st->c);
#ifdef CS_ALSA_MIDI
	if(st->midiseqh)
	{
		snd_seq_close(st->midiseqh);
		st->midiseqh = NULL;
	}
#endif
	memset(st, 0, sizeof(CS_state));
	st->running = 0;
#if 0
printf("CS_object:\t%d\n", sizeof(CS_object));
printf("CS_wave:\t%d\n", sizeof(CS_wave));
printf("CS_bank:\t%d\n", sizeof(CS_bank));
printf("CS_program:\t%d\n", sizeof(CS_program));
printf("CS_stackentry:\t%d\n", sizeof(CS_stackentry));
printf("CS_voice:\t%d\n", sizeof(CS_voice));
#endif
}
