/*
 * csapi.c - ChipSound asynchronous API implementation
 *
 * Copyright (C) 2010-2012 David Olofson <david@olofson.net>
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from the
 * use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software
 *    in a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 * 3. This notice may not be removed or altered from any source distribution.
 */

#include <stdlib.h>
#include <stdint.h>
#include <math.h>
#include "csinternals.h"

/*
static inline void MyLockAudio(void)
{
	printf("Lock! (%u)\n", SDL_GetTicks());
	SDL_LockAudio();
}
#define	SDL_LockAudio	MyLockAudio
*/

/*---------------------------------------------------------
	Async API message gateway
---------------------------------------------------------*/

CS_errors cs_OpenAPI(CS_state *st, float buffer)
{
	int i;
	int nmessages = CS_MINAPIMESSAGES + buffer * CS_TIMEAPIMESSAGES;
	if(sfifo_init(&st->fromapi, nmessages * sizeof(CS_messagebody)) < 0)
	{
		fprintf(stderr, "ChipSound: Could not open async API!\n");
		return CS_MEMORY;
	}
	nmessages = CS_MINMESSAGES + buffer * CS_TIMEMESSAGES;
	for(i = 0; i < nmessages; ++i)
	{
		CS_message *m = malloc(sizeof(CS_message));
		if(!m)
			return CS_MEMORY;
		m->next = st->messagepool;
		st->messagepool = m;
	}
	return CS_OK;
}


void cs_CloseAPI(CS_state *st)
{
	sfifo_close(&st->fromapi);
	while(st->messagepool)
	{
		CS_message *m = st->messagepool;
		st->messagepool = m->next;
		free(m);
	}
}


CS_errors cs_PumpMessages(CS_state *st)
{
	while(sfifo_used(&st->fromapi) >= sizeof(CS_messagebody))
	{
		CS_message *msg = cs_AllocMessage(st);
		if(!msg)
			return CS_MEMORY;
		if(sfifo_read(&st->fromapi, &msg->m,
				(unsigned)sizeof(CS_messagebody)) < 0)
		{
			fprintf(stderr, "ChipSound: Failed to read supposedly "
					"existing API message!\n");
			cs_FreeMessage(st, msg);
			return CS_INTERNAL + 10;
		}
		if(!st->groups[msg->m.gid])
		{
			cs_FreeMessage(st, msg);
			return CS_BADGROUP;
		}
		/* Adjust late messages, to avoid "bad" timestamps internally! */
		if(cs_TSDiff(msg->m.timestamp, st->now_frames) < 0)
			msg->m.timestamp = st->now_frames;
		cs_SendMessage(st->groups[msg->m.gid], msg);
	}
	return CS_OK;
}


/*---------------------------------------------------------
	Utilities
---------------------------------------------------------*/

float cs_F2P(float f)
{
	return log2(f / CS_MIDDLEC);
}


float cs_Rand(float max)
{
	CS_state *st = &cs_globalstate;
	return cs_Noise(st) * max / 65536.0f;
}


/*---------------------------------------------------------
	Playing and controlling
---------------------------------------------------------*/

void cs_Now(void)
{
	unsigned nt, nf;
	int dt;
	CS_state *st = &cs_globalstate;
	do {
		nf = st->now_frames;
		nt = nf + (st->audiospec.samples << 8);
		dt = SDL_GetTicks() - st->now_ticks;
	} while(nf != st->now_guard);
	if(dt < 0)
		dt = 0;	/* Audio has been off for a looooong time... */
	nt += (int64_t)st->msdur * dt >> 8;
//printf("cs_Now() ==> %f (was %f)\n", nt / 256.0f, st->timestamp / 256.0f);
	if(cs_TSDiff(nt, st->timestamp) >= 0)
		st->timestamp = nt;
//else printf("cs_Now(): Time went BACKWARDS!?\n");
}


void cs_Wait(float dt)
{
	CS_state *st = &cs_globalstate;
	st->timestamp += (unsigned)(st->msdur * dt / 256.0f);
//printf("cs_Wait(%f) ==> %f\n", dt, st->timestamp / 256.0f);
}


CS_errors cs_Starta(CS_gid gid, CS_vid vid, unsigned program,
		unsigned argc, int *argv)
{
	CS_state *st = &cs_globalstate;
	CS_messagebody msg;
	if(!st->running)
		return CS_NOTRUNNING;
	if(argc > CS_MAXARGS)
		return CS_MANYARGS;
	if(!(st->flags & CS_TIMESTAMP))
		cs_Now();
	msg.timestamp = st->timestamp;
	msg.addr = program;
	msg.vid = vid;
	msg.gid = gid;
	msg.what = CSM_STARTSUB;
	msg.argc = argc;
	memcpy(&msg.a, argv, sizeof(unsigned) * argc);
	if(sfifo_space(&st->fromapi) < sizeof(msg))
		return CS_OVERFLOW;
	if(sfifo_write(&st->fromapi, &msg, sizeof(msg)) != sizeof(msg))
		return CS_INTERNAL + 11;
//printf("%f: start\n", st->timestamp / 256.0f);
	return CS_OK;
}


CS_errors cs_Senda(CS_gid gid, CS_vid vid, unsigned h, unsigned argc, int *argv)
{
	CS_state *st = &cs_globalstate;
	CS_messagebody msg;
	if(!st->running)
		return CS_NOTRUNNING;
	if(argc > CS_MAXARGS)
		return CS_MANYARGS;
	if(!(st->flags & CS_TIMESTAMP))
		cs_Now();
	msg.timestamp = st->timestamp;
	msg.addr = h;
	msg.vid = vid;
	msg.gid = gid;
	msg.what = CSM_SENDSUB;
	msg.argc = argc;
	memcpy(&msg.a, argv, sizeof(unsigned) * argc);
	if(sfifo_space(&st->fromapi) < sizeof(msg))
		return CS_OVERFLOW;
	if(sfifo_write(&st->fromapi, &msg, sizeof(msg)) != sizeof(msg))
		return CS_INTERNAL + 12;
//printf("%f: send\n", st->timestamp / 256.0f);
	return CS_OK;
}


CS_errors cs_Detach(CS_gid gid, CS_vid vid)
{
	CS_state *st = &cs_globalstate;
	CS_messagebody msg;
	if(!st->running)
		return CS_NOTRUNNING;
	if(!(st->flags & CS_TIMESTAMP))
		cs_Now();
	msg.timestamp = st->timestamp;
	msg.vid = vid;
	msg.gid = gid;
	msg.what = CSM_DETACHSUB;
	msg.argc = 0;
	if(sfifo_space(&st->fromapi) < sizeof(msg))
		return CS_OVERFLOW;
	if(sfifo_write(&st->fromapi, &msg, sizeof(msg)) != sizeof(msg))
		return CS_INTERNAL + 13;
//printf("%f: detach\n", st->timestamp / 256.0f);
	return CS_OK;
}


static CS_errors csr_Kill(CS_state *st, CS_gid gid, CS_vid vid)
{
	int res = CS_NOTFOUND;
	CS_voice *v;
	if(!st->groups[gid])
		return CS_BADGROUP;
	for(v = st->groups[gid]->sub; v; v = v->next)
		if(!vid || (v->id == vid))
		{
			res = CS_OK;
			cs_VoiceKill(st, v);
		}
	return res;
}

CS_errors cs_Kill(CS_gid gid, CS_vid vid)
{
	int res;
	if(!cs_globalstate.running)
		return CS_NOTRUNNING;
	SDL_LockAudio();
	res = csr_Kill(&cs_globalstate, gid, vid);
	SDL_UnlockAudio();
	return res;
}


CS_errors cs_NewGroup(CS_gid parent, CS_gid gid)
{
	CS_state *st = &cs_globalstate;
	int res;
	if(!gid || (parent == gid) || !st->groups[parent])
		return CS_BADGROUP;
	if(st->groups[gid])
		cs_DetachGroup(gid);
	SDL_LockAudio();
	if(!(st->groups[gid] = cs_VoiceNew(st, st->groups[parent])))
	{
		SDL_UnlockAudio();
		return CS_VOICEALLOC;
	}
	st->groups[gid]->id = 0;	/* "Illegal" but not detached! */
	st->groups[gid]->r[R_SAMPLITUDE] = 65536;
	if((res = cs_VoiceCall(st, st->groups[gid], st->groupprog, 0, 0, NULL)))
		cs_VoiceFree(st, &st->groups[parent]->sub);
	SDL_UnlockAudio();
	return res;
}


CS_errors cs_SendGroupa(CS_gid gid, unsigned h, unsigned argc, int *argv)
{
	CS_state *st = &cs_globalstate;
	CS_messagebody msg;
	if(!st->running)
		return CS_NOTRUNNING;
	if(argc > CS_MAXARGS)
		return CS_MANYARGS;
	if(!(st->flags & CS_TIMESTAMP))
		cs_Now();
	msg.timestamp = st->timestamp;
	msg.addr = h;
	msg.gid = gid;
	msg.what = CSM_MESSAGE;
	msg.argc = argc;
	memcpy(&msg.a, argv, sizeof(unsigned) * argc);
	if(sfifo_space(&st->fromapi) < sizeof(msg))
		return CS_OVERFLOW;
	if(sfifo_write(&st->fromapi, &msg, sizeof(msg)) != sizeof(msg))
		return CS_INTERNAL + 14;
//printf("%f: sendgroup\n", st->timestamp / 256.0f);
	return CS_OK;
}


CS_errors cs_DetachGroup(CS_gid gid)
{
	CS_state *st = &cs_globalstate;
	if(!st->groups[gid])
		return CS_BADGROUP;
	cs_Detach(gid, 0);
/*
FIXME: This part can actually reach the voices before any pending messages,
FIXME: including those from the cs_Detach() above! This might cause problems in
FIXME: some situations...
*/
	SDL_LockAudio();
	if(gid)	/* The root group cannot be detached! */
	{
		cs_VoiceDetach(st->groups[gid]);
		st->groups[gid] = NULL;
	}
	SDL_UnlockAudio();
	return CS_OK;
}


/*---------------------------------------------------------
	Plugin interface
---------------------------------------------------------*/

CS_errors cs_Insert(CS_gid gid, CS_insert_cb callback, void *userdata)
{
	CS_state *st = &cs_globalstate;
	if(!st->groups[gid])
		return CS_BADGROUP;
	SDL_LockAudio();
	st->groups[gid]->insert_cb = callback;
	st->groups[gid]->insert_ud = userdata;
	SDL_UnlockAudio();
	return CS_OK;
}


/*---------------------------------------------------------
	Object property interface
---------------------------------------------------------*/

int cs_GetGlobalProp(CS_properties prop)
{
	CS_state *st = &cs_globalstate;
	int res;
	switch(prop)
	{
	  case CS_PSAMPLERATE:
		return st->audiospec.freq;
	  case CS_PBUFFER:
		return st->audiospec.samples;
	  case CS_PCHANNELS:
		return st->audiospec.channels;
	  case CS_PACTIVEVOICES:
		return st->activevoices;
	  case CS_PFREEVOICES:
		return st->totalvoices - st->activevoices;
	  case CS_PTOTALVOICES:
		return st->totalvoices;
	  case CS_PCPULOADAVG:
		if(!st->cputimecount)
			return 0;
		SDL_LockAudio();
		res = st->cputimesum / st->cputimecount;
		if(res)
			res = (st->now_ticks - st->avgtick) * 100 / res;
		if(res > 100)
			res = 100;
		SDL_UnlockAudio();
		return res;
	  case CS_PCPULOADMAX:
		return st->cpuloadmax;
	  case CS_PCPUTIMEAVG:
		if(!st->cputimecount)
			return 0;
		SDL_LockAudio();
		res = st->cputimesum / st->cputimecount;
		SDL_UnlockAudio();
		return res;
	  case CS_PCPUTIMEMAX:
		return st->cputimemax;
	  case CS_PINSTRUCTIONS:
		return st->instructions;
	  case CS_PEXPORTALL:
		return st->c->exportall;
	  default:
		return 0;
	}
}

CS_errors cs_SetGlobalProp(CS_properties prop, int v)
{
	CS_state *st = &cs_globalstate;
	switch(prop)
	{
	  case CS_PCPULOADAVG:
	  case CS_PCPUTIMEAVG:
		SDL_LockAudio();
		st->cputimesum = st->cputimecount = 0;
		st->avgtick = st->now_ticks;
		SDL_UnlockAudio();
		return CS_OK;
	  case CS_PCPULOADMAX:
		st->cpuloadmax = v;
		return CS_OK;
	  case CS_PCPUTIMEMAX:
		st->cputimemax = v;
		return CS_OK;
	  case CS_PINSTRUCTIONS:
		st->instructions = v;
		return CS_OK;
	  case CS_PEXPORTALL:
		st->c->exportall = v;
		return CS_OK;
	  default:
		return CS_NOTFOUND;
	}
}

int cs_GetObjectProp(unsigned handle, CS_properties prop)
{
	return 0;
}

int cs_GetGroupProp(CS_gid gid, CS_properties prop)
{
	return 0;
}

int cs_GetVoiceProp(CS_gid gid, CS_vid vid, CS_properties prop)
{
	return 0;
}
