/*
 * csbank.c - ChipSound banks and symbols
 *
 * Copyright (C) 2010-2012 David Olofson <david@olofson.net>
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from the
 * use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software
 *    in a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 * 3. This notice may not be removed or altered from any source distribution.
 */

#include <stdlib.h>
#include <stdint.h>
#include "csinternals.h"


/*---------------------------------------------------------
	Symbols
---------------------------------------------------------*/

CS_symbol *cs_NewSymbol(const char *name, CS_tokens token, int value)
{
	CS_symbol *s = (CS_symbol *)malloc(sizeof(CS_symbol));
	if(!s)
		return NULL;
	s->next = NULL;
	s->name = strdup(name);
	s->exported = 0;
	s->token = token;
	s->value = value;
	return s;
}

void cs_FreeSymbol(CS_symbol *s)
{
	free(s->name);
	free(s);
}

void cs_DeleteSymbol(CS_symbol **stack, CS_symbol *s)
{
	while(*stack != s)
		stack = &((*stack)->next);
	*stack = s->next;
	cs_FreeSymbol(s);
}

void cs_PushSymbol(CS_symbol **stack, CS_symbol *s)
{
	s->next = *stack;
	*stack = s;
}

CS_symbol *cs_FindSymbol(CS_symbol *s, const char *name)
{
	CS_state *st = &cs_globalstate;
	for( ; s; s = s->next)
		if(s->token == TK_BANK)
		{
			CS_bank *b = (CS_bank *)cs_ObjGet(st, s->value >> 16,
					CS_TBANK);
			CS_symbol *bs = cs_FindSymbol(b->exports, name);
			if(bs)
				return bs;
		}
		else if(!strcmp(name, s->name))
			return s;
	return NULL;
}


/*---------------------------------------------------------
	Object/handle management
---------------------------------------------------------*/

/* Allocate new object. Returns object handle, or a negative error code. */
int cs_ObjAlloc(CS_state *st, CS_otypes type)
{
	int size, oi = st->ffobject;
	switch(type)
	{
	  case CS_TBANK:	size = sizeof(CS_bank);		break;
	  case CS_TWAVE:	size = sizeof(CS_wave);		break;
	  case CS_TPROGRAM:	size = sizeof(CS_program);	break;
	  case CS_TUNIT:	size = sizeof(CS_unit);		break;
	  default:		return -(CS_INTERNAL + 3);
	}
	if(oi >= st->maxobjects)
	{
		int ns = oi ? oi * 3 / 2 : 16;
		CS_object **no, **oo;
		no = (CS_object **)malloc(ns * sizeof(void *));
		if(!no)
			return -CS_MEMORY;
		oi = st->maxobjects;
		memcpy(no, st->objects, oi * sizeof(void *));
		memset(no + oi + 1, 0, (ns - oi - 1) * sizeof(void *));
		oo = st->objects;
		SDL_LockAudio();
		st->objects = no;
		st->maxobjects = ns;
		SDL_UnlockAudio();
		free(oo);
	}
	st->objects[oi] = (CS_object *)calloc(1, size);
	if(!st->objects[oi])
		return -CS_MEMORY;
	st->objects[oi]->type = type;
	while((st->ffobject < st->maxobjects) && st->objects[st->ffobject])
		++st->ffobject;
	return oi;
}


CS_object *cs_ObjGet(CS_state *st, unsigned handle, CS_otypes type)
{
	if((handle < st->maxobjects) && st->objects[handle] &&
			st->objects[handle]->type == type)
		return st->objects[handle];
	return NULL;
}


void cs_ObjFree(CS_state *st, unsigned handle)
{
	int i;
	CS_object *o;
	if((handle > st->maxobjects) || !st->objects[handle])
		return;
	o = st->objects[handle];
	switch(o->type)
	{
	  case CS_TBANK:
	  {
		CS_bank *b = (CS_bank *)o;
		if(--b->refs)
			return;
		free(b->code);
		while(b->exports)
		{
			CS_symbol *s = b->exports;
			b->exports = s->next;
			switch(s->token)
			{
			  case TK_BANK:
			  case TK_PROGRAM:
			  case TK_WAVE:
				cs_ObjFree(st, s->value >> 16);
			  default:
				break;
			}
			cs_FreeSymbol(s);
		}
		break;
	  }
	  case CS_TWAVE:
	  {
		CS_wave *w = (CS_wave *)o;
		for(i = 0; i < CS_MIPLEVELS; ++i)
			free(w->data[i]);
		break;
	  }
	  case CS_TPROGRAM:
	  case CS_TUNIT:
		break;
	}
	free(o);
	st->objects[handle] = NULL;
	if(handle < st->ffobject)
		st->ffobject = handle;
}


/*---------------------------------------------------------
	Bank management
---------------------------------------------------------*/

int cs_NewBank(const char *name)
{
	CS_state *st = &cs_globalstate;
	char buf[64];
	CS_symbol *s;
	int bi = cs_ObjAlloc(st, CS_TBANK);
	CS_bank *bank = (CS_bank *)cs_ObjGet(st, bi, CS_TBANK);
	if(bi < 0)
		return bi;
	bank->refs = 1;
	if(!name)
	{
		snprintf(buf, sizeof(buf) - 1, "bank%p", bank);
		name = buf;
	}
	if(!(s = cs_NewSymbol(name, TK_BANK, bi << 16)))
	{
		cs_ObjFree(st, bi);
		return -CS_MEMORY;
	}
	cs_PushSymbol(&st->banks, s);
	return bi;
}

/* Kill any voices running code from bank 'b' */
static void cs_UntangleBank(CS_state *st, CS_voice *v, CS_bank *b)
{
	for( ; v; v = v->next)
	{
		CS_stackentry *se;
		if(v->program->bank == b)
		{
			cs_VoiceKill(st, v);
			continue;
		}
		for(se = v->stack; se; se = se->prev)
			if(se->program->bank == b)
			{
				cs_VoiceKill(st, v);
				continue;
			}
		if(v->sub)
			cs_UntangleBank(st, v->sub, b);
	}
}

/* Kill any voices playing non-existant waves */
static void cs_CheckWaves(CS_state *st, CS_voice *v)
{
	for( ; v; v = v->next)
	{
		cs_OscWave(&v->osc, st, v->r[R_WAVE] >> 16);
		if(!cs_ObjGet(st, v->r[R_WAVE] >> 16, CS_TWAVE))
			cs_VoiceKill(st, v);
		else if(v->sub)
			cs_CheckWaves(st, v->sub);
	}
}

void cs_FreeBank(int id)
{
	CS_state *st = &cs_globalstate;
	CS_symbol *s;
	CS_bank *b;
	if(id < 0)
	{
		int h;
		for(h = 1; h < st->maxobjects; ++h)
			if(cs_ObjGet(st, h, CS_TBANK))
				cs_FreeBank(h);
		return;
	}
	for(s = st->banks; s; s = s->next)
		if(s->value >> 16 == id)
			break;
	if(!s)
		return;
	cs_DeleteSymbol(&st->banks, s);
	b = (CS_bank *)cs_ObjGet(st, id, CS_TBANK);
	SDL_LockAudio();
	cs_UntangleBank(st, st->groups[0], b);
	for(s = b->exports; s; s = s->next)
		if(s->token == TK_WAVE)
			cs_ObjFree(st, s->value >> 16);
	cs_ObjFree(st, id);
	cs_CheckWaves(st, st->groups[0]);
	SDL_UnlockAudio();
}


/*---------------------------------------------------------
	Loading sounds
---------------------------------------------------------*/

int cs_LoadString(const char *code, const char *name)
{
	CS_state *st = &cs_globalstate;
	int bi = cs_NewBank(name), res;
	CS_bank *bank;
	if(bi < 0)
		return bi;
	bank = (CS_bank *)cs_ObjGet(st, bi, CS_TBANK);
	if((res = cs_CompileString(st->c, bank, code, name)) != CS_OK)
		return -res;
	return bi;
}

int cs_Load(const char *fn)
{
	CS_state *st = &cs_globalstate;
	int bi = cs_NewBank(fn), res;
	CS_bank *bank;
	if(bi < 0)
		return bi;
	bank = (CS_bank *)cs_ObjGet(st, bi, CS_TBANK);
	if((res = cs_CompileFile(st->c, bank, fn)) != CS_OK)
		return -res;
	return bi;
}


/*---------------------------------------------------------
	Waveform management
---------------------------------------------------------*/

static int cs_SampleSize(CS_sampleformats fmt)
{
	switch(fmt)
	{
	  case CS_I8:			return 1;
	  case CS_I16:			return 2;
	  case CS_I32:	case CS_F32:	return 4;
	  default:			return 0;
	}
}


static int cs_WaveAlloc(CS_bank *bank, const char *name, unsigned length,
		unsigned period, int flags)
{
	CS_state *st = &cs_globalstate;
	int i, wave = cs_ObjAlloc(st, CS_TWAVE);
	CS_wave *w = (CS_wave *)cs_ObjGet(st, wave, CS_TWAVE);
	CS_symbol *s;
	if(wave < 0)
		return wave;
	w->period = period;
	w->looped = flags & CS_LOOPED;
	w->size[0] = length;
	for(i = 0; i < CS_MIPLEVELS; ++i)
	{
		int size = (w->size[0] + (1 << i) - 1) >> i;
		w->size[i] = size;
		size = CS_WAVEPRE + size + CS_WAVEPOST;
		w->data[i] = (int16_t *)malloc(size * sizeof(int16_t));
		if(!w->data[i])
		{
			cs_ObjFree(st, wave);
			return -1;
		}
	}
	if(!(s = cs_NewSymbol(name, TK_WAVE, wave << 16)))
	{
		cs_ObjFree(st, wave);
		return -CS_MEMORY;
	}
	cs_PushSymbol(&bank->exports, s);
	w->Process = length ? cso_Wavetable : cso_Off;
	return wave;
}


static void cs_WaveFixPad(CS_wave *w, unsigned miplevel)
{
	int16_t *d = w->data[miplevel];
	unsigned size = w->size[miplevel];
	if(w->looped && size)
	{
		int i;
		memcpy(d, d + size, CS_WAVEPRE * 2);
		for(i = 0; i < CS_WAVEPOST; ++i)
			d[CS_WAVEPRE + size + i] = d[CS_WAVEPRE + i % size];
	}
	else
	{
		memset(d, 0, CS_WAVEPRE * 2);
		memset(d + CS_WAVEPRE + size, 0, CS_WAVEPOST * 2);
	}
}


static CS_errors cs_WaveUpload2(CS_wave *w,
		CS_sampleformats fmt, const void *data, unsigned length)
{
	int i, s;
	int16_t *d = w->data[0] + CS_WAVEPRE;
	switch(fmt)
	{
	  case CS_I8:
		for(s = 0; s < length; ++s)
			d[s] = ((int8_t *)data)[s] << 8;
		break;
	  case CS_I16:
		memcpy(d, data, length * 2);
		break;
	  case CS_I32:
		for(s = 0; s < length; ++s)
			d[s] = ((int32_t *)data)[s] >> 16;
		break;
	  case CS_F32:
		for(s = 0; s < length; ++s)
			d[s] = ((float *)data)[s] * 32767.0f;
		break;
	}
	cs_WaveFixPad(w, 0);
	for(i = 1; i < CS_MIPLEVELS; ++i)
	{
		int16_t *sd = w->data[i - 1] + CS_WAVEPRE;
		d = w->data[i] + CS_WAVEPRE;
		for(s = 0; s < w->size[i]; ++s)
			d[s] = (((int)sd[s * 2] << 1) + sd[s * 2 - 1] +
					sd[s * 2 + 1]) >> 2;
		cs_WaveFixPad(w, i);
	}
#if 0
	printf("%s: -------------\n", w->name);
	for(i = 0; i < CS_MIPLEVELS; ++i)
	{
		d = w->data[i] + CS_WAVEPRE;
		printf(" %d:\n", i);
		printf("(");
		for(s = 0; s < CS_WAVEPRE; ++s)
			printf("\t%d", d[s - CS_WAVEPRE]);
		printf("\t)\n");
		for(s = 0; s < w->size[i]; ++s)
			printf("\t%d", d[s]);
		printf("\n");
		printf("(");
		for(s = 0; s < CS_WAVEPOST; ++s)
			printf("\t%d", d[w->size[i] + s]);
		printf("\t)\n");
	}
#endif
	return CS_OK;
}


int cs_UploadWave(unsigned bank, const char *name, unsigned period, int flags,
		CS_sampleformats fmt, const void *data, unsigned size)
{
	CS_state *st = &cs_globalstate;
	CS_wave *wave;
	CS_bank *b = (CS_bank *)cs_ObjGet(st, bank, CS_TBANK);
	int ss = cs_SampleSize(fmt), w, res;
	if(!b)
		return -CS_BADBANK;
	if(!ss)
		return -CS_BADFORMAT;
	if((w = cs_WaveAlloc(b, name, size / ss, period, flags)) < 0)
		return w;
	wave = (CS_wave *)cs_ObjGet(st, w, CS_TWAVE);
	if((res = cs_WaveUpload2(wave, fmt, data, size / ss)) != CS_OK)
	{
		cs_ObjFree(st, w);
		return -res;
	}
	return w;
}


int cs_Oscillator(unsigned bank, const char *name, unsigned period,
		CS_osccb process)
{
	CS_state *st = &cs_globalstate;
	int w;
	CS_wave *wave;
	CS_bank *b = (CS_bank *)cs_ObjGet(st, bank, CS_TBANK);
	if(!b)
		return -CS_BADBANK;
	if((w = cs_WaveAlloc(b, name, 0, period, 1)) < 0)
		return w;
	wave = (CS_wave *)cs_ObjGet(st, w, CS_TWAVE);
	wave->Process = process;
	return w;
}


/*---------------------------------------------------------
	Exports
---------------------------------------------------------*/

static int cs_GetBank(unsigned bank)
{
	CS_state *st = &cs_globalstate;
	CS_symbol *s;
	for(s = st->banks; s; s = s->next, --bank)
		if(!bank)
			return s->value >> 16;
	return -CS_BADBANK;
}

static CS_symbol *cs_GetSymbol(int bank, unsigned i)
{
	CS_state *st = &cs_globalstate;
	int b0, b1;
	if(bank == -1)
		b0 = 0, b1 = 100000;
	else
		b0 = b1 = bank;
	for(bank = b0; bank <= b1; ++bank)
	{
		CS_symbol *s;
		CS_bank *b = (CS_bank *)cs_ObjGet(st, cs_GetBank(bank), CS_TBANK);
		if(!b)
			return NULL;
		for(s = b->exports; s; s = s->next, --i)
			if(!i)
				return s;
	}
	return NULL;
}

int cs_Find(int bank, const char *name)
{
	CS_state *st = &cs_globalstate;
	int h;
	CS_bank *b;
	CS_symbol *s;
	if(bank == -1)
		for(bank = 0; 1; ++bank)
			if((h = cs_Find(cs_GetBank(bank), name)) == -CS_BADBANK)
				return -CS_NOTFOUND;
			else if(h >= 0)
				return h;
	if(!(b = (CS_bank *)cs_ObjGet(st, bank, CS_TBANK)))
		return -CS_BADBANK;
	if(!(s = cs_FindSymbol(b->exports, name)))
		return -CS_NOTFOUND;
	return s->value >> 16;
}

int cs_Get(int bank, unsigned i)
{
	CS_symbol *s = cs_GetSymbol(bank, i);
	return s ? s->value >> 16 : -CS_NOTFOUND;
}

const char *cs_GetName(int bank, unsigned i)
{
	CS_symbol *s = cs_GetSymbol(bank, i);
	return s ? s->name : "<not found>";
}


int cs_GetType(unsigned handle)
{
	CS_state *st = &cs_globalstate;
	if((handle > st->maxobjects) || !st->objects[handle])
		return -1;
	return st->objects[handle]->type;
}


const char *cs_TypeName(int type)
{
	switch(type)
	{
	  case CS_TBANK:	return "bank";
	  case CS_TPROGRAM:	return "program";
	  case CS_TWAVE:	return "wave";
	  default:		return "<unknown type>";
	}
}
