/*
 * csc.c - ChipSound compiler
 *
 * Copyright (C) 2010-2012 David Olofson <david@olofson.net>
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from the
 * use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software
 *    in a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 * 3. This notice may not be removed or altered from any source distribution.
 *
WARNING: The ChipSound parser uses setjmp()/longjmp() for error handling.
WARNING: Calls with the csc_ prefix MUST ONLY be used with cs_Try()!
 */

/* Define to dump compiled banks to "out.csb" */
#undef	CS_OUT_CSB

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "csinternals.h"


/*
 * Grab 'n' characters from s, and append a 'nul' byte.
 * (NOT equivalent to strndup())
 */
static inline char *csc_ndup(const char *s, size_t n)
{
	char *new = malloc(n + 1);
	if(!new)
		return NULL;
	new[n] = '\0';
	return memcpy(new, s, n);
}


/*
 * Issue VM instruction 'op' with arguments 'reg' and 'arg'.
 * Argument range checking is done, and 'arg' is treated as integer or 16:16
 * and handled apropriately, based on the opcode.
 */
static void csc_Code(CS_cstate *c, unsigned op, unsigned reg, int arg)
{
	CS_bank *bank = c->bank;
	if(c->nocode)
		csc_Throw(c, CS_NOCODE);
	if(bank->end + 2 >= bank->size)
	{
		int i, ns = bank->size;
		unsigned *nc;
		if(ns)
			while(bank->end + 2 >= ns)
				ns = ns * 3 / 2;
		else
			ns = 1024;
		SDL_LockAudio();
		nc = (unsigned *)realloc(bank->code, ns * sizeof(unsigned));
		if(!nc)
		{
			SDL_UnlockAudio();
			csc_Throw(c, CS_MEMORY);
		}
		bank->code = nc;
		bank->size = ns;
		SDL_UnlockAudio();
		for(i = bank->end; i < bank->size; ++i)
			bank->code[i] = OP_END << 26;
	}
	if(op >= CS_OPCODES)
		csc_Throw(c, CS_BADOPCODE);
	if(reg >= CS_REGISTERS)
		csc_Throw(c, CS_BADREGISTER);
	switch(op)
	{
	  case OP_JUMP:
	  case OP_WAKE:
	  case OP_FORCE:
	  case OP_LOOP:
	  case OP_JZ:
	  case OP_JNZ:
	  case OP_JG:
	  case OP_JL:
	  case OP_JGE:
	  case OP_JLE:
		if(arg < 0)
			csc_Throw(c, CS_BADJUMP);
		if(arg == bank->end)
			csc_Throw(c, CS_INFLOOP);
		if(arg > bank->end)
			csc_Throw(c, CS_BADJUMP);
		break;
	  case OP_RUN:
	  case OP_SPAWN:
	  case OP_SPAWND:
		if(!cs_ObjGet(c->enginestate, arg, CS_TPROGRAM))
			csc_Throw(c, CS_BADPROGRAM);
		break;
	  case OP_SEND:
	  case OP_SENDA:
	  case OP_SENDS:
		if(arg > CS_MAXEPS)
			csc_Throw(c, CS_BADENTRY);
		break;
	  case OP_LOADR:
		if(arg == reg)
			return;	/* NOP! */
	  case OP_ADDR:
	  case OP_SUBR:
	  case OP_MULR:
	  case OP_DIVR:
	  case OP_MODR:
	  case OP_RANDR:
	  case OP_P2DR:
	  case OP_RUNR:
	  case OP_QUANTR:
		if(arg > CS_REGISTERS)
			csc_Throw(c, CS_BADREG2);
		break;
	  default:
		arg = cs_i2f(arg);
		break;
	}
	if(arg > 0x1fffff)
		csc_Throw(c, CS_BADIMMARG);
	bank->code[bank->end++] = (op << 26) | (reg << 21) | arg;
DBG(	cs_DumpIns(bank->code, bank->end - 1);)
}


static CS_errors csc_DoFixups(CS_cstate *c, CS_symbol *s)
{
	CS_bank *bank = c->bank;
	if(s->value > 0x1fffff)
		return CS_BADIMMARG;
	while(s->fixups)
	{
		CS_fixup *fx = s->fixups;
		s->fixups = fx->next;
		bank->code[fx->pos] |= s->value;
DBG(	printf("FIXUP: "); cs_DumpIns(bank->code, fx->pos);)
		free(fx);
	}
	return CS_OK;
}


static inline int cs_GetChar(CS_cstate *c)
{
	int ch = c->code[c->l.codepos];
	if(!ch)
		return -1;
	++c->l.codepos;
	return ch;
}

/*
 * Read decimal value, return as signed 16:16 fixed point.
 * Returns CS_OK if everything is normal, otherwise an appropriate error code;
 *	CS_OVERFLOW if the parsed value is too large to fit in a 16:16 fixp
 *	CS_UNDERFLOW if a non-zero parsed value is truncated to zero
 *	CS_BADVALUE if a proper number could not be parsed at all
 */
static CS_errors cs_GetNum(CS_cstate *c, int ch, int *v)
{
	int sign = 1;
	double val = 0.0f;
	unsigned xp = 0;
	int valid = 0;	/* At least one figure, or this is not a number! */
	if(ch == '-')
	{
		sign = -1;
		ch = cs_GetChar(c);
	}
	for(;; ch = cs_GetChar(c))
		if((ch >= '0') && (ch <= '9'))
		{
			xp *= 10;
			val *= 10.0f;
			val += ch - '0';
			valid = 1;
		}
		else if(ch == '.')
		{
			if(xp)
			{
				--c->l.codepos;
				*v = 0;
				return CS_BADVALUE;
			}
			xp = 1;
		}
		else
		{
			if(!valid)
			{
				--c->l.codepos;
				*v = 0;
				return CS_BADVALUE;
			}
			val *= sign;
			if(xp)
				val /= xp;
			if(ch == 'n')
				val /= 12.0f;
			else if(ch == 'f')
				val = cs_F2P(val);
			else
				--c->l.codepos;
			if(val >= 32767.0f)
				*v = 0x7fffffff;
			else if(val <= -32768.0f)
				*v = 0x80000000;
			else
			{
				*v = val * 65536.0f + 0.5f;
				return val && !*v ? CS_UNDERFLOW : CS_OK;
			}
			return CS_OVERFLOW;
		}
}

/* NOTE:
 *	This returns names as new symbols. Use cs_Grab() to keep one of those -
 *	or it will be deleted as parsing continues!
 *	   Also note that TK_NAME symbols MUST NOT be used in any symbol tables,
 *	as the parser may delete them whenever they end up in *l.sym!
 */
static CS_tokens csc_Lex(CS_cstate *c)
{
	char *name;
	int nstart, ch, prevch;
	if(c->pl.sym && (c->pl.sym->token == TK_NAME))
		cs_FreeSymbol(c->pl.sym);	/* Free if not grabbed! */
	c->pl = c->l;
	c->l.sym = NULL;
	while(1)
	{
		switch((ch = cs_GetChar(c)))
		{
		  case -1:
			return (c->l.token = TK_EOF);
		  case ',':
		  case ';':
		  case '\n':
			c->l.val = ch;
			return (c->l.token = TK_EOS);
		  case ' ':
		  case '\t':
		  case '\r':
			continue;
		  case '/':
			switch(cs_GetChar(c))
			{
			  case '/':
				while((ch = cs_GetChar(c)) != '\n')
					if(ch == -1)
						return (c->l.token = TK_EOF);
				--c->l.codepos;
				continue;
			  case '*':
				for(prevch = 0; (ch = cs_GetChar(c)) != -1;
						prevch = ch)
					if((prevch == '*') && (ch == '/'))
						break;
				continue;
			}
			--c->l.codepos;
			break;
		}
		break;
	}
	if(((ch >= '0') && (ch <= '9')) || (ch == '-') || (ch == '.'))
	{
		CS_errors res = cs_GetNum(c, ch, &c->l.val);
		if(res == CS_OK)
			return (c->l.token = TK_VALUE);
		else if(res != CS_BADVALUE)
			csc_Throw(c, res);
	}
	nstart = c->l.codepos - 1;
	while(1)
	{
		if(((ch >= 'a') && (ch <= 'z')) || ((ch >= 'A') && (ch <= 'Z'))
				|| ((ch >= '0') && (ch <= '9')))
		{
			ch = cs_GetChar(c);
			continue;
		}
		if(nstart == c->l.codepos - 1)
			return (c->l.token = ch);
		--c->l.codepos;
		name = csc_ndup(c->code + nstart, c->l.codepos - nstart);
		c->l.sym = cs_FindSymbol(c->symbols, name);
		if(!c->l.sym)
		{
			c->l.sym = cs_NewSymbol(name, TK_NAME, 0);
			if(!c->l.sym)
			{
				free(name);
				csc_Throw(c, CS_MEMORY);
			}
		}
		c->l.val = c->l.sym->value;
		free(name);
		return (c->l.token = c->l.sym->token);
	}
}

static void csc_Unlex(CS_cstate *c)
{
	if(c->l.sym && (c->l.sym->token == TK_NAME))
		cs_FreeSymbol(c->l.sym);
	c->l = c->pl;
	memset(&c->pl, 0, sizeof(CS_lexstate));
}

static void csc_SkipLF(CS_cstate *c)
{
	while(csc_Lex(c) == TK_EOS)
		if(c->l.val != '\n')
			break;
	csc_Unlex(c);
}

static CS_symbol *csc_Grab(CS_cstate *c, CS_symbol *s)
{
	if(c->l.sym == s)
		c->l.sym = NULL;
	if(c->pl.sym == s)
		c->pl.sym = NULL;
	return s;
}


typedef struct CS_scope
{
	CS_symbol	*symbols;
	unsigned	regtop;
} CS_scope;

static void csc_BeginScope(CS_cstate *c, CS_scope *sc)
{
	sc->symbols = c->symbols;
	sc->regtop = c->regtop;
}

static void csc_EndScope(CS_cstate *c, CS_scope *sc)
{
	int res = CS_OK;
	c->regtop = sc->regtop;
	c->l.sym = c->pl.sym = NULL;
	while(c->symbols != sc->symbols)
	{
		CS_symbol *s = c->symbols;
		c->symbols = s->next;
		if(s->token == TK_FWDECL)
			res = CS_UNDEFSYM;
		if(s->exported)
			cs_PushSymbol(&c->bank->exports, s);
		else
			cs_FreeSymbol(s);
	}
	if(res)
		csc_Throw(c, res);
}

static unsigned csc_AllocReg(CS_cstate *c)
{
	if(c->regtop >= CS_REGISTERS)
		csc_Throw(c, CS_OUTOFREGS);
	return c->regtop++;
}

static void csc_FreeReg(CS_cstate *c, unsigned r)
{
#ifdef DEBUG
	if(r != c->regtop - 1)
	{
		fprintf(stderr, "ChipSound INTERNAL ERROR: Tried to free VM"
				"registers out of order!\n");
		csc_Throw(c, CS_INTERNAL + 100);
	}
#endif
	--c->regtop;
}


static void csc_Expect(CS_cstate *c, CS_tokens tk, CS_errors err)
{
	if(csc_Lex(c) != tk)
		csc_Throw(c, err);
}


static void csc_Branch(CS_cstate *c, CS_opcodes op, int xtk, int xval,
		unsigned to, int *fixpos)
{
	int r;
	if(cs_IsValue(xtk))
	{
		/* FIXME:
		 *	These are no conditionals! They should be evaluated
		 *	at compile time instead.
		 */
		r = csc_AllocReg(c);
		csc_Code(c, OP_LOAD, r, xval);
		if(fixpos)
			*fixpos = c->bank->end;
		csc_Code(c, op, r, to);
		csc_FreeReg(c, r);
	}
	else if(cs_IsRegister(xtk))
	{
		if(fixpos)
			*fixpos = c->bank->end;
		csc_Code(c, op, xval, to);
		if(xtk == TK_TEMPREG)
			csc_FreeReg(c, xval);
	}
	else
		csc_Throw(c, CS_INTERNAL + 101);
}


static void csc_VarDecl(CS_cstate *c, CS_symbol *s)
{
	s->token = TK_REGISTER;
	s->value = csc_AllocReg(c);
	cs_PushSymbol(&c->symbols, s);
}


static void csc_CodeOpR(CS_cstate *c, CS_opcodes op, int to, unsigned r)
{
	switch(op)
	{
	  case OP_ADD:
	  case OP_MUL:
	  case OP_MOD:
	  case OP_QUANT:
	  case OP_RAND:
	  case OP_LOAD:
		csc_Code(c, op + 1, to, r);
		break;
	  case OP_DELAY:
	  case OP_TDELAY:
	  case OP_PHASE:
	  case OP_DEBUG:
		csc_Code(c, op + 1, r, 0);
		break;
	  case OP_SUBR:
	  case OP_DIVR:
	  case OP_P2DR:
		csc_Code(c, op, to, r);
		break;
	  default:
		csc_Throw(c, CS_BADOPCODE);
	}
}

static void csc_CodeOpV(CS_cstate *c, CS_opcodes op, int to, int v)
{
	int tmpr = to;
	switch(op)
	{
	  case OP_MOD:
		if(!v)
			csc_Throw(c, CS_DIVBYZERO);
	  case OP_ADD:
	  case OP_MUL:
	  case OP_QUANT:
	  case OP_RAND:
	  case OP_LOAD:
	  case OP_DELAY:	/* ('to' is not used by these last four) */
	  case OP_TDELAY:
	  case OP_PHASE:
	  case OP_DEBUG:
		csc_Code(c, op, to, v);
		break;
	  case OP_SUBR:
		csc_Code(c, OP_ADD, to, -v);
		break;
	  case OP_DIVR:
		if(!v)
			csc_Throw(c, CS_DIVBYZERO);
		csc_Code(c, OP_MUL, to, 0x100000000LL / v);
		break;
	  default:
		if((op != OP_RAND) && (op != OP_P2DR))
			tmpr = csc_AllocReg(c);
		csc_Code(c, OP_LOAD, tmpr, v);
		csc_CodeOpR(c, op, to, tmpr);
		if(tmpr != to)
			csc_FreeReg(c, tmpr);
		break;
	}
}

/*
 * Issue code '<op> <to> <from>', where <from> is derived from the parser state
 * after a csc_SimplExp() call. Any temporary register will be freed.
 */
static void csc_CodeOpL(CS_cstate *c, CS_opcodes op, int to)
{
	switch(c->l.token)
	{
	  case TK_VALUE:
	  case TK_WAVE:
	  case TK_PROGRAM:
		csc_CodeOpV(c, op, to, c->l.val);
		break;
	  case TK_TEMPREG:
	  case TK_REGISTER:
		csc_CodeOpR(c, op, to, c->l.val);
		if(c->l.token == TK_TEMPREG)
			csc_FreeReg(c, c->l.val);
		break;
	  default:
		csc_Throw(c, CS_INTERNAL + 102);	/* We should never get here! */
	}
}


static void csc_SimplExp(CS_cstate *c, int r);

/* Expression, terminated with ')' */
static void csc_Expression(CS_cstate *c, int to)
{
	int op;
	csc_SkipLF(c);
	csc_SimplExp(c, to);
	if((c->l.token == TK_PROGRAM) || (c->l.token == TK_WAVE))
		csc_Throw(c, CS_NEXPHANDLE); /* No arithmetics on these! */
	csc_CodeOpL(c, OP_LOAD, to);
	while(1)
	{
		csc_SkipLF(c);
		switch(csc_Lex(c))
		{
		  case ')':
			return;	/* Done !*/
		  case TK_INSTRUCTION:	op = c->l.val;	break;
		  case '+':		op = OP_ADD;	break;
		  case '-':		op = OP_SUBR;	break;
		  case '*':		op = OP_MUL;	break;
		  case '/':		op = OP_DIVR;	break;
		  case '%':		op = OP_MOD;	break;
		  default:		csc_Throw(c, CS_EXPOP);
		}
		csc_SkipLF(c);
		csc_SimplExp(c, -1);
		if((c->l.token == TK_PROGRAM) || (c->l.token == TK_WAVE))
			csc_Throw(c, CS_NEXPHANDLE);
		csc_CodeOpL(c, op, to);
	}
}


/*
 * Parse value or variable, or issue code to evaluate expression. Returns with
 * the result as the current token.
 *
 * If the expression issues code that can target a register, 'r' is used.
 *
 * If 'r' is -1, a temporary register is allocated, and the token is set to
 * TK_TEMPREG. In this case, the caller is responsible for freeing the
 * temporary register!
 */
static void csc_SimplExp(CS_cstate *c, int r)
{
	int op, tmpr = r;
	switch(csc_Lex(c))
	{
	  case TK_VALUE:
	  case TK_WAVE:
	  case TK_PROGRAM:
	  case TK_REGISTER:
		return;		/* Done! Just return as is - no code! */
	  case '(':
		if(r < 0)
			tmpr = csc_AllocReg(c);
		csc_Expression(c, tmpr);
		break;
	  case TK_INSTRUCTION:	/* Unary operator */
		op = c->l.val;
		if((op != OP_P2DR) && (op != OP_RAND))
			csc_Throw(c, CS_NOTUNARY);
		if(r < 0)
			tmpr = csc_AllocReg(c);
		csc_SimplExp(c, tmpr);
		csc_CodeOpL(c, op, tmpr);
		break;
	  default:
		csc_Throw(c, CS_EXPVALREG);
	}
	c->l.token = r < 0 ? TK_TEMPREG : TK_REGISTER; /* Temp or allocated? */
	c->l.val = tmpr;
}


static void csc_Arguments(CS_cstate *c, int maxargc)
{
	int argc;
	for(argc = 0; argc <= maxargc; ++argc)
	{
		if((csc_Lex(c) == '}') || (c->l.token == TK_EOS))
		{
			csc_Unlex(c);
			return;		/* Done! */
		}
		csc_Unlex(c);
		csc_SimplExp(c, -1);
		if(cs_IsValue(c->l.token))
			csc_Code(c, OP_PUSH, 0, c->l.val);
		else if(cs_IsRegister(c->l.token))
		{
			csc_Code(c, OP_PUSHR, c->l.val, 0);
			if(c->l.token == TK_TEMPREG)
				csc_FreeReg(c, c->l.val);
		}
	}
	csc_Throw(c, CS_MANYARGS);
}


static void csc_Instruction(CS_cstate *c, CS_opcodes op, int r)
{
	int i, p;
	switch(op)
	{
	  case OP_END:
		csc_Code(c, op, 0, 0);
		break;
	  case OP_WAKE:
	  case OP_FORCE:
		if(!c->inhandler)
			csc_Throw(c, CS_NOWAKEFORCE);
	  case OP_JUMP:
		if((csc_Lex(c) != TK_LABEL) && (c->l.token != TK_FWDECL))
			csc_Throw(c, CS_EXPLABEL);
		csc_Code(c, op, 0, c->l.val);
		break;
	  case OP_LOOP:
		csc_Expect(c, TK_REGISTER, CS_EXPVARIABLE);
		r = c->l.val;
		csc_Expect(c, TK_LABEL, CS_EXPLABEL);
		csc_Code(c, op, r, c->l.val);
		break;
	  case OP_JZ:
	  case OP_JNZ:
	  case OP_JG:
	  case OP_JL:
	  case OP_JGE:
	  case OP_JLE:
		csc_SimplExp(c, -1);
		csc_Expect(c, TK_LABEL, CS_EXPLABEL);
		csc_Branch(c, op, c->pl.token, c->pl.val, c->l.val, NULL);
		break;
	  case OP_RUN:
		if(c->inhandler)
			csc_Throw(c, CS_NORUN);
		/* Fall through! */
	  case OP_SPAWN:
	  case OP_SPAWND:
		switch(csc_Lex(c))
		{
		  case TK_REGISTER:
			++op;
			p = c->l.val;
			i = CS_MAXARGS;	/* Can't check these compile time... */
			break;
		  case TK_PROGRAM:
			p = c->l.val >> 16;	/* Program handle */
			i = ((CS_program *)cs_ObjGet(c->enginestate, p,
					CS_TPROGRAM))->eps[0].argc;
			break;
		  default:
			csc_Throw(c, CS_EXPPROGRAM);
		}
		csc_Arguments(c, i);
		csc_Code(c, op, r, p);
		break;
	  case OP_SEND:
	  case OP_SENDS:
	  case OP_SENDA:
		csc_Expect(c, TK_VALUE, CS_EXPVALUE);
		p = c->l.val >> 16;	/* Message handler entry point */
		if(!p)
			csc_Throw(c, CS_BADENTRY); /* 0 is not for messages! */
		csc_Arguments(c, CS_MAXARGS);
		csc_Code(c, op, r, p);
		break;
	  case OP_KILL:
/*TODO: Expression!!! */
		if(csc_Lex(c) == '*')
			csc_Code(c, OP_KILLA, 0, 0);
		else
		{
			if(c->l.token != TK_VALUE)
				csc_Throw(c, CS_EXPVALUE);
			csc_Code(c, OP_KILL, c->l.val >> 16, 0);
		}
		break;
	  case OP_SET:
		csc_Expect(c, TK_REGISTER, CS_EXPVARIABLE);
		csc_Code(c, OP_SET, c->l.val, 0);
		break;
	  case OP_DELAY:
	  case OP_TDELAY:
		if(c->inhandler)
			csc_Throw(c, CS_NOTIMING);
		/* Fall through! */
	  case OP_PHASE:
	  case OP_DEBUG:
		csc_SimplExp(c, -1);
		csc_CodeOpL(c, op, 0);
		break;
	  case OP_ADD:
	  case OP_SUBR:
	  case OP_MUL:
	  case OP_DIVR:
	  case OP_MOD:
	  case OP_QUANT:
	  case OP_RAND:
	  case OP_P2DR:
		if(csc_Lex(c) == '!')
		{
			CS_symbol *s;
			if((op != OP_RAND) && (op != OP_P2DR))
				csc_Throw(c, CS_BADVARDECL);
			csc_Expect(c, TK_NAME, CS_EXPNAME);
			s = csc_Grab(c, c->l.sym);
			csc_VarDecl(c, s);
			r = s->value;
		}
		else
		{
			if(c->l.token != TK_REGISTER)
				csc_Throw(c, CS_EXPVARIABLE);
			r = c->l.sym->value;
		}
		csc_SimplExp(c, (op == OP_RAND) || (op == OP_P2DR) ? r : -1);
		csc_CodeOpL(c, op, r);
		break;
	  default:
		csc_Throw(c, CS_BADOPCODE);
	}
}


static void csc_Def(CS_cstate *c)
{
	CS_symbol *s;
	csc_Expect(c, TK_NAME, CS_EXPNAME);
	s = csc_Grab(c, c->l.sym);
	switch(csc_Lex(c))
	{
	  case TK_VALUE:
		s->token = TK_VALUE;
		s->value = c->l.val;
		break;
	  case TK_LABEL:
	  case TK_REGISTER:
	  case TK_WAVE:
	  case TK_PROGRAM:
		s->token = c->l.sym->token;
		s->value = c->l.sym->value;
		break;
	  default:
		csc_Throw(c, CS_BADVALUE);
	}
	cs_PushSymbol(&c->symbols, s);
}


static void csc_Body(CS_cstate *c, CS_tokens terminator);


static void csc_ArgList(CS_cstate *c, unsigned ep)
{
	unsigned *argc = &c->program->eps[ep].argc;
	c->program->eps[ep].argv = c->regtop;
	for(*argc = 0; csc_SkipLF(c), csc_Lex(c) != ')'; ++*argc)
	{
		if(*argc > CS_MAXARGS)
			csc_Throw(c, CS_MANYARGS);
		if(c->l.token != TK_NAME)
			csc_Throw(c, CS_EXPNAME);
		csc_VarDecl(c, csc_Grab(c, c->l.sym));
		if(csc_Lex(c) == '=')
		{
			if(!cs_IsValue(csc_Lex(c)))
				csc_Throw(c, CS_EXPVALUE);
			c->program->eps[ep].argdefs[*argc] = c->l.val;
		}
		else
			csc_Unlex(c);
	}
}


static void csc_ProgDef(CS_cstate *c, CS_symbol *s, int export)
{
	int i;
	CS_scope sc;
	if(s->token != TK_NAME)
		csc_Throw(c, CS_EXPNAME); /* TODO: Forward declarations. */
	if(c->program || c->inhandler)
		csc_Throw(c, CS_NOPROGHERE);
	s->token = TK_PROGRAM;
	if((s->value = cs_ObjAlloc(c->enginestate, CS_TPROGRAM)) < 0)
		csc_Throw(c, -s->value);
	c->program = (CS_program *)cs_ObjGet(c->enginestate, s->value, CS_TPROGRAM);
	s->value <<= 16;
	c->program->bank = c->bank;
	c->program->eps[0].position = c->bank->end;
	for(i = 1; i < CS_MAXEPS; ++i)
		c->program->eps[i].position = 0xffffffff;
	s->exported = export || c->exportall;
	cs_PushSymbol(&c->symbols, s);
DBG(	printf("program %s():\n", s->name);)
	csc_BeginScope(c, &sc);
	csc_ArgList(c, 0);
	csc_SkipLF(c);
	csc_Expect(c, '{', CS_EXPBODY);
	c->inhandler = c->nocode = 0;
	csc_Body(c, '}');
	if(!c->nocode)
		csc_Code(c, OP_END, 0, 0);
	csc_EndScope(c, &sc);
	c->program = NULL;
	c->nocode = 1;
}


static void csc_MsgDef(CS_cstate *c, unsigned ep)
{
	CS_scope sc;
	if(ep >= CS_MAXEPS)
		csc_Throw(c, CS_BADENTRY);
	if(!c->program || c->inhandler)
		csc_Throw(c, CS_NOMSGHERE);
	if(!c->nocode)
		csc_Code(c, OP_END, 0, 0);
	c->program->eps[ep].position = c->bank->end;
DBG(	printf("message %d():\n", ep);)
	csc_BeginScope(c, &sc);
	csc_ArgList(c, ep);
	csc_SkipLF(c);
	csc_Expect(c, '{', CS_EXPBODY);
	c->inhandler = 1;
	c->nocode = 0;
	csc_Body(c, '}');
	c->inhandler = 0;
	csc_Code(c, OP_END, 0, 0);
	csc_EndScope(c, &sc);
	c->nocode = 1;
}


static void csc_IfWhile(CS_cstate *c, CS_opcodes op, int loop)
{
	int fixpos, loopto = c->bank->end;
	csc_SimplExp(c, -1);
	csc_Branch(c, op, c->l.token, c->l.val, 0, &fixpos);
	csc_SkipLF(c);
	csc_Expect(c, '{', CS_EXPBODY);
	csc_Body(c, '}');
	if(csc_Lex(c) == TK_ELSE)
	{
		int fixelse = c->bank->end;
		if(loop)
			csc_Throw(c, CS_NEXPELSE);
		csc_Code(c, OP_JUMP, 0, 0);	/* To skip over 'else' body */
		if(fixpos >= 0)		/* False condition lands here! */
		{
			c->bank->code[fixpos] |= c->bank->end;
DBG(	printf("FIXUP: "); cs_DumpIns(c->bank->code, fixpos);)
		}
		csc_SkipLF(c);
		csc_Expect(c, '{', CS_EXPBODY);
		csc_Body(c, '}');
		c->bank->code[fixelse] |= c->bank->end;
DBG(	printf("FIXUP: "); cs_DumpIns(c->bank->code, fixelse);)
		return;
	}
	else
		csc_Unlex(c);
	if(loop)
		csc_Code(c, OP_JUMP, 0, loopto);
	if(fixpos >= 0)
	{
		c->bank->code[fixpos] |= c->bank->end;
DBG(	printf("FIXUP: "); cs_DumpIns(c->bank->code, fixpos);)
	}
}


static void csc_TimesL(CS_cstate *c)
{
	int loopto, r = csc_AllocReg(c);
	csc_CodeOpL(c, OP_LOAD, r);
	loopto = c->bank->end;
	csc_Expect(c, '{', CS_EXPBODY);
	csc_Body(c, '}');
	csc_Code(c, OP_LOOP, r, loopto);
	csc_FreeReg(c, r);
}


static void csc_For(CS_cstate *c)
{
	int loopto = c->bank->end;
	csc_SkipLF(c);
	csc_Expect(c, '{', CS_EXPBODY);
	csc_Body(c, '}');
	csc_Code(c, OP_JUMP, 0, loopto);
}


static int csc_Statement(CS_cstate *c, CS_tokens terminator)
{
	int r;
	switch(csc_Lex(c))
	{
	  case TK_VALUE:
/*TODO: Do this for expressions as well, for '<', ':' and '{' */
		r = c->l.val >> 16;
		switch(csc_Lex(c))
		{
		  case '(':
			csc_MsgDef(c, r);
			return 1;
		  case '{':
			csc_Unlex(c);
			csc_TimesL(c);
			return 1;
		  case '<':
			csc_Instruction(c, OP_SEND, r);
			break;
		  case ':':
			csc_Instruction(c, OP_SPAWN, r);
			break;
		  default:
			csc_Throw(c, CS_NEXPVALUE);
		}
		break;
	  case TK_REGISTER:
		if(csc_Lex(c) == '{')
		{
			csc_Unlex(c);
			csc_TimesL(c);
		}
		else
		{
			csc_Unlex(c);
			r = c->l.val;
			csc_SimplExp(c, r);
			csc_CodeOpL(c, OP_LOAD, r);
		}
		break;
	  case '.':		/* Label or local program */
		if((csc_Lex(c) != TK_NAME) && (c->l.token != TK_FWDECL))
			csc_Throw(c, CS_BADLABEL);
		if(csc_Lex(c) == '(')
			csc_ProgDef(c, csc_Grab(c, c->pl.sym), 0);
		else
		{
			CS_symbol *s;
			csc_Unlex(c);
			s = csc_Grab(c, c->l.sym);
			s->token = TK_LABEL;
			s->value = c->bank->end;
			cs_PushSymbol(&c->symbols, s);
DBG(	printf("label .%s:\n", s->name);)
			if(c->l.token == TK_FWDECL)
				csc_DoFixups(c, s);
		}
		return 1;
	  case TK_FWDECL:	/* TODO: Program forward declarations. */
		csc_Throw(c, CS_SYMBOLDEF);
	  case TK_NAME:
		if(csc_Lex(c) != '(')
			csc_Throw(c, CS_NEXPNAME);
		csc_ProgDef(c, csc_Grab(c, c->pl.sym), 1);
		break;
	  case TK_LABEL:
		csc_Throw(c, CS_SYMBOLDEF);	/* Already defined! */
	  case '!':
	  {
		CS_symbol *s;
		switch(csc_Lex(c))	/* For nicer error messages... */
		{
		  case TK_NAME:
			break;
		  case TK_REGISTER:
		  case TK_LABEL:
		  case TK_PROGRAM:
			csc_Throw(c, CS_SYMBOLDEF);
		  default:
			csc_Throw(c, CS_EXPNAME);
		}
		s = csc_Grab(c, c->l.sym);
		csc_VarDecl(c, s);
		csc_SimplExp(c, s->value);
		csc_CodeOpL(c, OP_LOAD, s->value);
		break;
	  }
	  case ':':
		csc_Instruction(c, OP_SPAWND, 0);
		break;
	  case '<':
		csc_Instruction(c, OP_SENDS, 0);
		break;
	  case '+':
		csc_Instruction(c, OP_ADD, 0);
		break;
	  case '-':
		csc_Instruction(c, OP_SUBR, 0);
		break;
	  case '*':
		if(csc_Lex(c) == '<')
			csc_Instruction(c, OP_SENDA, 0);
		else
		{
			csc_Unlex(c);
			csc_Instruction(c, OP_MUL, 0);
		}
		break;
	  case '/':
		csc_Instruction(c, OP_DIVR, 0);
		break;
	  case '%':
		csc_Instruction(c, OP_MOD, 0);
		break;
	  case TK_INSTRUCTION:
		csc_Instruction(c, c->l.val, 0);
		break;
	  case TK_PROGRAM:
		csc_Unlex(c);
		csc_Instruction(c, OP_SPAWND, 0);
		break;
	  case TK_TEMPO:
		/* Calculate (1000 / (<tempo> / 60 * <tbp>)) */
		r = csc_AllocReg(c);
		csc_SimplExp(c, r);
		csc_CodeOpL(c, OP_LOAD, r);
		csc_Code(c, OP_MUL, r, 65536 / 60);
		csc_SimplExp(c, r);
		csc_CodeOpL(c, OP_MUL, r);
		csc_Code(c, OP_LOAD, R_TICK, 1000 << 16);
		csc_Code(c, OP_DIVR, R_TICK, r);
		csc_FreeReg(c, r);
		break;
	  case TK_DEF:
		csc_Def(c);
		return 1;
	  case TK_IF:
		csc_IfWhile(c, c->l.val, 0);
		return 1;
	  case TK_WHILE:
		csc_IfWhile(c, c->l.val, 1);
		return 1;
	  case TK_FOR:
		csc_For(c);
		return 1;
	  case '{':
		csc_Body(c, '}');
		return 1;
	  case TK_EOS:
		return 1;
	  default:
		if(c->l.token != terminator)
			csc_Throw(c, CS_NEXPTOKEN);
		return 0;
	}
	/* Finalizer for statements that expect a terminator */
	if(csc_Lex(c) == TK_EOS)
		return 1;
	if(c->l.token != terminator)
		csc_Throw(c, CS_EXPEOS);
	return 0;
}


static void csc_Body(CS_cstate *c, CS_tokens terminator)
{
	CS_scope sc;
	csc_BeginScope(c, &sc);
	while(csc_Statement(c, terminator))
		;
	csc_EndScope(c, &sc);
}


static struct
{
	const char	*n;
	CS_tokens	tk;
	int		v;
} initsyms [] = {
	/* Hardwired "root" bank 0 */
	{"root", TK_BANK, 0},

	/* Hardwired voice control registers */
	{"tick", TK_REGISTER, R_TICK},
	{"w", TK_REGISTER, R_WAVE},
	{"p", TK_REGISTER, R_PITCH},
	{"sp", TK_REGISTER, R_SPITCH},
	{"a", TK_REGISTER, R_AMPLITUDE},
	{"sa", TK_REGISTER, R_SAMPLITUDE},
	{"vol", TK_REGISTER, R_VOLUME},
	{"pan", TK_REGISTER, R_PANX},

	/* Instructions */
	{"end", TK_INSTRUCTION, OP_END},
	{"jump", TK_INSTRUCTION, OP_JUMP},
	{"jz", TK_INSTRUCTION, OP_JZ},
	{"jnz", TK_INSTRUCTION, OP_JNZ},
	{"jg", TK_INSTRUCTION, OP_JG},
	{"jl", TK_INSTRUCTION, OP_JL},
	{"jge", TK_INSTRUCTION, OP_JGE},
	{"jle", TK_INSTRUCTION, OP_JLE},
	{"wake", TK_INSTRUCTION, OP_WAKE},
	{"force", TK_INSTRUCTION, OP_FORCE},
	{"loop", TK_INSTRUCTION, OP_LOOP},
	{"run", TK_INSTRUCTION, OP_RUN},
	{"kill", TK_INSTRUCTION, OP_KILL},
	{"d", TK_INSTRUCTION, OP_DELAY},
	{"td", TK_INSTRUCTION, OP_TDELAY},
	{"phase", TK_INSTRUCTION, OP_PHASE},
	{"quant", TK_INSTRUCTION, OP_QUANT},
	{"rand", TK_INSTRUCTION, OP_RAND},
	{"p2d", TK_INSTRUCTION, OP_P2DR},
	{"set", TK_INSTRUCTION, OP_SET},
	{"debug", TK_INSTRUCTION, OP_DEBUG},

	/* Directives, macros, keywords... */
	{"def", TK_DEF, 0},
//	{"uses", TK_USES, 0},
	{"tempo", TK_TEMPO, 0},
	{"if", TK_IF, OP_JZ},
	{"ifz", TK_IF, OP_JNZ},
	{"ifl", TK_IF, OP_JG},
	{"ifg", TK_IF, OP_JL},
	{"ifle", TK_IF, OP_JGE},
	{"ifge", TK_IF, OP_JLE},
	{"else", TK_ELSE, 0},
	{"while", TK_WHILE, OP_JZ},
	{"wz", TK_WHILE, OP_JNZ},
	{"wl", TK_WHILE, OP_JGE},
	{"wg", TK_WHILE, OP_JLE},
	{"wle", TK_WHILE, OP_JG},
	{"wge", TK_WHILE, OP_JL},
	{"for", TK_FOR, 0},

	{NULL, 0, 0}
};


CS_errors cs_OpenCompiler(CS_state *st, int flags)
{
	int i;
	CS_cstate *c = (CS_cstate *)calloc(1, sizeof(CS_cstate));
	if(!c)
		return CS_MEMORY;
	c->enginestate = st;
	st->c = c;
	c->regtop = CS_CREGISTERS;	/* Control regs are hardwired! */
	c->exportall = (flags & CS_EXPORTALL) == CS_EXPORTALL;
	for(i = 0; initsyms[i].n; ++i)
	{
		CS_symbol *s = cs_NewSymbol(initsyms[i].n, initsyms[i].tk,
				initsyms[i].v);
		if(!s)
			return CS_MEMORY;
		cs_PushSymbol(&c->symbols, s);
	}
	return CS_OK;
}


void cs_CloseCompiler(CS_cstate *c)
{
	while(c->symbols)
	{
		CS_symbol *s = c->symbols;
		c->symbols = s->next;
		cs_FreeSymbol(s);
	}
}


static void cs_Compile(CS_cstate *c, CS_scope *sc, const char *source)
{
	csc_Try(c)
	{
		csc_BeginScope(c, sc);
		csc_Body(c, TK_EOF);
		csc_EndScope(c, sc);
	}
	csc_Except
	{
		int line, col, i;
		for(line = col = 1, i = 0; i < c->l.codepos ; ++i)
			if(c->code[i] == '\n')
				++line, col = 1;
			else
				col += (c->code[i] == '\t') ? 8 : 1;
		fprintf(stderr, "ChipSound: %s at line %d, column %d in "
				"\"%s\"\n",
				cs_ErrorString(c->error), line, col, source);
		/* Try to avoid dangling wires and stuff... */
		csc_Try(c)
		{
			csc_Code(c, OP_END, 0, 0);
		}
		csc_Except
		{
			fprintf(stderr, "111: %s\n", cs_ErrorString(c->error));
		}
		csc_Try(c)
		{
			csc_EndScope(c, sc);
		}
		csc_Except
		{
			fprintf(stderr, "222: %s\n", cs_ErrorString(c->error));
		}
	}
}


CS_errors cs_CompileString(CS_cstate *c, CS_bank *bank, const char *code,
		const char *source)
{
	CS_scope sc;
	c->bank = bank;
	c->code = code;
	c->l.codepos = 0;
	c->program = NULL;
	c->inhandler = 0;
	c->nocode = 1;
	cs_Compile(c, &sc, source);
#if 0
	csc_Try(c)
	{
		c->bank = bank;
		c->code = code;
		c->l.codepos = 0;
		c->program = NULL;
		c->inhandler = 0;
		c->nocode = 1;
		csc_BeginScope(c, &sc);
		csc_Body(c, TK_EOF);
		csc_EndScope(c, &sc);
	}
	csc_Except
	{
		int line, col, i;
		for(line = col = 1, i = 0; i < c->l.codepos ; ++i)
			if(code[i] == '\n')
				++line, col = 1;
			else
				col += (code[i] == '\t') ? 8 : 1;
		fprintf(stderr, "ChipSound: %s at line %d, column %d in "
				"\"%s\"\n",
				cs_ErrorString(c->error), line, col, source);
		csc_Try(c)
		{	/* Try to avoid dangling wires and stuff... */
			csc_Code(c, OP_END, 0, 0);
			csc_EndScope(c, &sc);
		}
	}
#endif
	return c->error;
}


CS_errors cs_CompileFile(CS_cstate *c, CS_bank *bank, const char *fn)
{
	int res;
	char *code;
	FILE *f;
	size_t fsize;
	f = fopen(fn, "rb");
	if(!f)
		return CS_OPEN;
	fseek(f, 0, SEEK_END);
	fsize = ftell(f);
	fseek(f, 0, SEEK_SET);
	code = malloc(fsize + 1);
	if(!code)
	{
		fclose(f);
		return CS_MEMORY;
	}
	code[fsize] = 0;
	if(fread(code, 1, fsize, f) != fsize)
	{
		fclose(f);
		free(code);
		return CS_READ;
	}
	fclose(f);
	res = cs_CompileString(c, bank, code, fn);
	free(code);
#ifdef CS_OUT_CSB
	{
		CS_symbol *s = bank->exports;
		f = fopen("out.csb", "wb");
		fwrite(&bank->end, 1, 3, f);
		fwrite(bank->code, bank->end, sizeof(unsigned), f);
		for( ; s; s = s->next)
		{
			fwrite(s->name, 1, strlen(s->name) + 1, f);
			fwrite(&s->value, 1, 3, f);
		}
		fclose(f);
	}
#endif
	return res;
}


/*---------------------------------------------------------
	Debug tools
---------------------------------------------------------*/

static const char *cs_errnames[100] = {
/*0*/	"Ok - no error!",
	"VM program ended normally",
	"Out of memory",
	"Buffer overflow",
	"Division by zero",
	"Error opening file",
	"Error opening device",
	"Bad file or device I/O format",
/*8*/	"Error reading file",
	"VM overload; too many instructions back-to-back",
	"Could not allocate voice",
	"ChipSound not initialized or running",
	"Object not found",
	"Invalid bank handle",
	"Invalid waveform handle",
	"Invalid program handle",
/*16*/	"Invalid program entry point",
	"Group does not exist, or cannot be used here",
	"Voice does not exist, or bad voice id",
	"Too many arguments to VM program",
	"Export already declared",
	"Expected value or register",
	"Expected end of statement",
	"Expected closing brace",
/*24*/	"Expected label",
	"Expected program",
	"Expected name",
	"Expected body",
	"Expected operator",
	"Bad label name",
	"Symbol already defined",
	"Undefined symbol",
/*32*/	"Expected variable",
	"Expected value",
	"Value not expected here",
	"Handle not expected here",
	"Bad value",
	"Value does not fit in numeric type",
	"Value too small; would truncate to zero",
	"Unexpected token",
/*40*/	"Unexpected end of file",
	"Illegal jump target position",
	"Jump would cause infinite loop",
	"Invalid VM opcode",
	"Invalid VM register index",
	"Invalid VM register index, second argument",
	"Immediate argument out of range",
	"Out of VM registers",
/*48*/	"Program cannot be declared here",
	"Message cannot be declared here",
	"Variable cannot be declared here",
	"Declared variable not initialized",
	"Not a unary operator",
	"Code not allowed here",
	"Timing instructions not allowed here",
	"Cannot run program from here",
/*56*/	"Undefined symbols in program",
	"'else' not applicable here",
	"'wake' and 'force' not applicable here",
	"Expected unit"
};

static char cs_errbuf[64];

const char *cs_ErrorString(unsigned errorcode)
{
	if(errorcode < CS_INTERNAL)
		return cs_errnames[errorcode];
	else
	{
		cs_errbuf[sizeof(cs_errbuf) - 1] = 0;
		snprintf(cs_errbuf, sizeof(cs_errbuf) - 1,
				"INTERNAL ERROR #%d; please report to "
				"<david@olofson.net>", errorcode - CS_INTERNAL);
		return cs_errbuf;
	}
}

static const char *cs_insnames[CS_OPCODES] = {
	"END",		"JUMP",		"LOOP",		"JZ",
	"JNZ",		"JG",		"JL",		"JGE",
	"JLE",		"WAKE",		"FORCE",	"SUBR",
	"DIVR",		"P2DR",		"DELAY",	"DELAYR",
	"TDELAY",	"TDELAYR",	"PHASE",	"PHASER",
	"LOAD",		"LOADR",	"ADD",		"ADDR",
	"MUL",		"MULR",		"MOD",		"MODR",
	"QUANT",	"QUANTR",	"RAND",		"RANDR",
	"PUSH",		"PUSHR",	"RUN",		"RUNR",
	"SPAWN",	"SPAWNR",	"SPAWND",	"SPAWNDR",
	"SEND",		"SENDA",	"SENDS",	"SET",
	"KILL",		"KILLA",	"DEBUG",	"DEBUGR"
};

static const char *cs_regnames[CS_CREGISTERS] = {
	"TICK",	"W",	"P",	"SP",	"A",	"SA",	"VOL",	"PAN"
};

static void cs_PrintRegName(unsigned r)
{
	if(r < CS_CREGISTERS)
		fputs(cs_regnames[r], stdout);
	else
		printf("R%d", r);
}

void cs_DumpIns(unsigned *code, unsigned pc)
{
	unsigned ins = code[pc];
	CS_opcodes op = ins >> 26;
	int reg = (ins >> 21) & 0x1f;
	int arg = ins & 0x1fffff;
	printf("%d:\t%-8.8s", pc, cs_insnames[op]);
	switch(op)
	{
	  /* No arguments */
	  case OP_END:
	  case OP_KILLA:
		break;
	  /* integer */
	  case OP_JUMP:
	  case OP_WAKE:
	  case OP_FORCE:
	  case OP_RUN:
	  case OP_SENDA:
	  case OP_SENDS:
		printf("%d", arg);
		break;
	  /* f20 */
	  case OP_DELAY:
	  case OP_TDELAY:
	  case OP_PHASE:
	  case OP_PUSH:
	  case OP_DEBUG:
		printf("%f", cs_f2i(arg) / 65536.0f);
		break;
	  /* register */
	  case OP_DELAYR:
	  case OP_TDELAYR:
	  case OP_PUSHR:
	  case OP_SET:
	  case OP_RUNR:
	  case OP_SPAWNDR:
	  case OP_PHASER:
	  case OP_DEBUGR:
		cs_PrintRegName(reg);
		break;
	  /* register f20 */
	  case OP_LOAD:
	  case OP_ADD:
	  case OP_MUL:
	  case OP_MOD:
	  case OP_QUANT:
	  case OP_RAND:
		cs_PrintRegName(reg);
		printf(" %f", cs_f2i(arg) / 65536.0f);
		break;
	  /* register integer */
	  case OP_LOOP:
	  case OP_JZ:
	  case OP_JNZ:
	  case OP_JG:
	  case OP_JL:
	  case OP_JGE:
	  case OP_JLE:
		cs_PrintRegName(reg);
		printf(" %d", arg);
		break;
	  /* index */
	  case OP_KILL:
		printf("%d", reg);
		break;
	  /* index integer */
	  case OP_SPAWN:
	  case OP_SPAWND:
	  case OP_SEND:
		printf("%d %d", reg, arg);
		break;
	  /* register register */
	  case OP_LOADR:
	  case OP_ADDR:
	  case OP_SUBR:
	  case OP_MULR:
	  case OP_DIVR:
	  case OP_MODR:
	  case OP_QUANTR:
	  case OP_RANDR:
	  case OP_SPAWNR:
	  case OP_P2DR:
		cs_PrintRegName(reg);
		printf(" ");
		cs_PrintRegName(arg);
		break;
	}
	printf("\n");
}
