/*
 * csinternals.h - ChipSound internals
 *
 * Copyright (C) 2010-2012 David Olofson <david@olofson.net>
 *
 * This software is provided 'as-is', without any express or implied warranty.
 * In no event will the authors be held liable for any damages arising from the
 * use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software
 *    in a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 * 3. This notice may not be removed or altered from any source distribution.
 *
WARNING: The ChipSound parser uses setjmp()/longjmp() for error handling.
WARNING: Calls with the csc_ prefix MUST ONLY be used with csc_Try()!
 */

#ifndef CSINTERNALS_H
#define CSINTERNALS_H

/*
 * API message FIFO size coefficients. Units are *messages* - not bytes!
 *
 * NOTE:
 *	The calculated FIFO size is a hard limit! The FIFO cannot be scaled up
 *	after initialization in the current implementation.
 *
 * CS_MINAPIMESSAGES is the minimum size regardless of audio buffering/latency;
 * effectively the maximum number of messages that can reliably be sent
 * back-to-back.
 *
 * CS_TIMEAPIMESSAGES in the number of additional messages per second to
 * allocate buffer space for. This is needed when using large audio buffers, as
 * the message FIFO is only checked once per audio buffer!
 */
#define	CS_MINAPIMESSAGES	256
#define	CS_TIMEAPIMESSAGES	1000

/*
 * Initial message pool size coefficients, corresponding to the above.
 *
 * NOTE:
 *	This is just the initial pool size. The pool will be extended
 *	automatically as needed, although that currently means memory
 *	allocation from the audio context, which is a very bad idea in
 *	a low latency realtime application!
 *
TODO: Increase initial pool size if using messages internally!
TODO: (As of now, they're only used for API->voice.)

 * Also note that CS_TIMEMESSAGES is mostly motivated by the API, as internal
 * voice->voice messago lifetime is governed by CS_MAXFRAG.
 */
#define	CS_MINMESSAGES		256
#define	CS_TIMEMESSAGES		1000

/*
 * Uncomment this to disable ALSA MIDI input support where ALSA is present.
*/
#undef	CS_ALSA_MIDI

#include "sfifo.h"
#include "csunits.h"
#include "a_limiter.h"
#include "SDL.h"
#ifdef CS_ALSA_MIDI
#	include <alsa/asoundlib.h>
#	define	MIDIGRP	(CS_MAXGIDS - 1)
#endif

#ifdef DEBUG
#	include	<stdio.h>
#	include	<assert.h>
#	define	DBG(x)		x	/* General debug output */
#	define	NUMMSGS(x)	x	/* Message order tracking */
#	define	DUMPMSGS(x)	x	/* Dump messages from the VM/msg loop */
#	define	DUMPCODE(x)		/* Enable realtime VM code dumping */
#	undef	CERRDIE			/* Die in assert(0) on compile errors */
#else
#	define	DBG(x)
#	define	NUMMSGS(x)
#	define	DUMPMSGS(x)
#	define	DUMPCODE(x)
#	undef	CERRDIE
#endif

/* Compiler error handling */
#include <setjmp.h>
#define	CS_jumpbuf	jmp_buf
#define	csc_Try(c)	if(!setjmp((c)->jumpbuf))
#define csc_Except	else
#ifdef CERRDIE
#	define	csc_Throw(c, x)	({					\
			printf("THROW %s\n", cs_ErrorString(c->error));	\
			assert(0);					\
		})
#else
#	define	csc_Throw(c, x)	longjmp((c)->jumpbuf, ((c)->error = (x)))
#endif

/* Enable high fidelity wavetable oscillator */
#define	CS_HIFI

/* Default tick duration; corresponds to 'tempo 120 4' */
#define	CS_DEFAULTTICK	(125 << 16)

/* Waveform period for full bandwidth down to 20 Hz fundamental */
#define	SC_WPER		2048

/*
 * Max number of VM instructions to run without processing samples. If exceeded,
 * the voice will be killed, to avoid freezing the whole sound engine.
 */
#define CS_INSLIMIT	10000

/* Initial pool sizes */
#define	CS_INITVOICES	256
#define	CS_INITBLOCKS	128

/*
 * Interpolators: Max number of samples required around the current sample. Must
 * also cover the mipmapping resampler, and waves shorter than CS_MAXFRAG!
 */
#define	CS_INTERPRE	1
#define	CS_INTERPOST	2

/* Interpolator: Maximum phase increment allowed at any mipmap level. (24:8) */
#define	CS_MAXPHINC	512

/* Number of samples to prepend before, and append to, any waveform. */
#define	CS_WAVEPRE	CS_INTERPRE
#define	CS_WAVEPOST	\
		(CS_INTERPOST + ((CS_MAXFRAG * (CS_MAXPHINC + 256)) >> 8) + 1)

typedef struct CS_object CS_object;
typedef struct CS_wave CS_wave;
typedef struct CS_bank CS_bank;
typedef struct CS_program CS_program;
typedef struct CS_oscillator CS_oscillator;
typedef struct CS_panmixer CS_panmixer;
typedef struct CS_message CS_message;
typedef struct CS_stackentry CS_stackentry;
typedef struct CS_voice CS_voice;
typedef struct CS_state CS_state;
typedef struct CS_symbol CS_symbol;
typedef struct CS_fixup CS_fixup;
typedef struct CS_cstate CS_cstate;

typedef void (*CS_osccb)(CS_oscillator *o, int *buffer, unsigned frames);

/*
 * Compare two timestamps. Returns (a - b) with correct handling of timestamp
 * wrapping.
 *
 * NOTE:
 *	This will assume that b is BEFORE a if the difference is more than half
 *	of the "wrap period"! This is to allow proper handling of messages that
 *	arrive late.
 */
static inline int cs_TSDiff(unsigned a, unsigned b)
{
	return (int)(a - b);
}


/*---------------------------------------------------------
	CSL Compiler
---------------------------------------------------------*/

typedef enum CS_tokens
{
	/* (lsym points to the matching symbol where applicable) */
	TK_EOF = 256,	/* End of file */
	TK_EOS,	/* End of statement */
	TK_VALUE,	/* lval = value (16:16 fixp); lsym = NULL */
	TK_TEMPREG,	/* Temporary register; lval = register index */
	TK_BANK,	/* lval = value (handle) */
	TK_WAVE,	/* lval = value (handle) */
	TK_PROGRAM,	/* lval = value (handle) */
	TK_NAME,	/* lsym = new symbol (add or free!) */
	TK_FWDECL,	/* lsym = symbol (accumulates fixups!) */
	TK_LABEL,	/* lval = value from symbol (code position) */
	TK_REGISTER,	/* lval = value from symbol (register index) */
	TK_INSTRUCTION,/* lval = value from symbol (pseudo opcode) */
	TK_DEF,		/* 'def' directive */
	TK_TEMPO,	/* 'tempo' macro "instruction" */
	TK_IF,		/* "if*" conditional variants */
	TK_ELSE,	/* 'else' keyword */
	TK_WHILE,	/* "while|w*" loop variants */
	TK_FOR		/* "for" loop */
} CS_tokens;

static inline int cs_IsValue(CS_tokens tk)
{
	return (tk == TK_VALUE) || (tk == TK_WAVE) || (tk == TK_PROGRAM);
}

static inline int cs_IsRegister(CS_tokens tk)
{
	return (tk == TK_TEMPREG) || (tk == TK_REGISTER);
}

/* Convert f20 to int32 */
static inline int cs_f2i(int f)
{
	return f << 16 >> (f >> 16);
}

/* Convert int32 to f20 */
static inline int cs_i2f(int i)
{
	int x = 0;
	int n = 0;
	if(i < 0)
		n = 1,	i = -i;
	if(!(i & 0x7f800000))
		x = 8,	i <<= 8;
	if(!(i & 0x78000000))
		x += 4,	i <<= 4;
	if(!(i & 0x60000000))
		x += 2,	i <<= 2;
	if(!(i & 0x40000000))
		x += 1,	i <<= 1;
	if(n)
		i = -i;
	return ((unsigned) i >> 16) | (x << 16);
}


/*---------------------------------------------------------
	Symbol tables/stacks
---------------------------------------------------------*/

struct CS_symbol
{
	CS_symbol	*next;
	char		*name;
	CS_fixup	*fixups;
	int		exported;
	CS_tokens	token;
	int		value;
};

/* Create new symbol, WITHOUT adding it to the table. */
CS_symbol *cs_NewSymbol(const char *name, CS_tokens token, int value);

/* Free symbol. MUST be detached from symbol table first! */
void cs_FreeSymbol(CS_symbol *s);

/* Detach and free symbol */
void cs_DeleteSymbol(CS_symbol **stack, CS_symbol *s);

/* Push symbol onto symbol stack. */
void cs_PushSymbol(CS_symbol **stack, CS_symbol *s);

/* Find symbol by name. */
CS_symbol *cs_FindSymbol(CS_symbol *s, const char *name);


/*---------------------------------------------------------
	VM and code generation
---------------------------------------------------------*/

/*
 * VM instructions. (Some of these are used by the parser as well.)
 * NOTE: *R versions MUST be right after their non-R counterparts!
 */
typedef enum CS_opcodes
{
	OP_END = 0,	OP_JUMP,	OP_LOOP,	OP_JZ,
	OP_JNZ,		OP_JG,		OP_JL,		OP_JGE,
	OP_JLE,		OP_WAKE,	OP_FORCE,	OP_SUBR,
	OP_DIVR,	OP_P2DR,	OP_DELAY,	OP_DELAYR,
	OP_TDELAY,	OP_TDELAYR,	OP_PHASE,	OP_PHASER,
	OP_LOAD,	OP_LOADR,	OP_ADD,		OP_ADDR,
	OP_MUL,		OP_MULR,	OP_MOD,		OP_MODR,
	OP_QUANT,	OP_QUANTR,	OP_RAND,	OP_RANDR,
	OP_PUSH,	OP_PUSHR,	OP_RUN,		OP_RUNR,
	OP_SPAWN,	OP_SPAWNR,	OP_SPAWND,	OP_SPAWNDR,
	OP_SEND,	OP_SENDA,	OP_SENDS,	OP_SET,
	OP_KILL,	OP_KILLA,	OP_DEBUG,	OP_DEBUGR
#define	CS_OPCODES	(OP_DEBUGR + 1)
} CS_opcodes;

typedef enum CS_cregisters
{
	R_TICK = 0,
	R_WAVE,
	R_PITCH,
	R_SPITCH,
	R_AMPLITUDE,
	R_SAMPLITUDE,
	R_VOLUME,
	R_PANX
#define	CS_CREGISTERS	(R_PANX + 1)
} CS_cregisters;

/* Compile ChipSound Language source code into VM code. */
CS_errors cs_CompileString(CS_cstate *c, CS_bank *bank, const char *code,
		const char *source);
CS_errors cs_CompileFile(CS_cstate *c, CS_bank *bank, const char *fn);

void cs_DumpIns(unsigned *code, unsigned pc);


/*---------------------------------------------------------
	8:24 control ramping device
---------------------------------------------------------*/

typedef struct CS_ramper
{
	int	value;		/* Current value */
	int	target;		/* Target value */
	int	delta;		/* Per-sample delta */
	int	frames;		/* Sample frames to end of ramp */
} CS_ramper;

/* Prepare ramper for some processing */
static inline void cs_PrepareRamp(CS_ramper *rr, int frames)
{
	if(!rr->frames)
	{
		rr->value = rr->target;
		rr->delta = 0;
	}
	else if(frames <= rr->frames)
	{
		rr->delta = (rr->target - rr->value) / rr->frames;
		rr->frames -= frames;
	}
	else
	{
		/*
		 * This isn't accurate, but the error is limited by CS_MAXFRAG,
		 * and it only ever happens with asynchronous ramps anyway!
		 */
		rr->delta = (rr->target - rr->value) / frames;
		rr->frames = 0;
	}
}

/* Advance ramper by 'frames' */
static inline void cs_RunRamp(CS_ramper *rr, int frames)
{
	rr->value += rr->delta * frames;
}

/* Set up ramp from current value to 'target' (16:16!) over 'frames' */
static inline void cs_SetRamp(CS_ramper *rr, int target, int frames)
{
	rr->frames = frames;
	rr->target = target << 8;
	if(!frames)
	{
		rr->delta = 0;
		rr->value = rr->target;
	}
}


/*---------------------------------------------------------
	Engine structures
---------------------------------------------------------*/

struct CS_object
{
	CS_otypes	type;
	/* (Implementation follows) */
};

/* CS_object: Waveform with mipmaps */
struct CS_wave
{
	CS_otypes	type;
	int		looped;			/* 1 if waveform is looped */
	unsigned	period;			/* Fundamental period length */
	int16_t		*data[CS_MIPLEVELS];	/* One buffer per mip level */
	unsigned	size[CS_MIPLEVELS];	/* Sizes EXCLUDING pre/post! */
	CS_osccb	Process;		/* Oscillator implementation */
};

/* CS_object: VM code bank */
struct CS_bank
{
	CS_otypes	type;
	unsigned	refs;		/* References to this bank */
	int		end;		/* First unused code position */
	int		size;		/* Size of 'code' */
	unsigned	*code;		/* VM code */
	CS_symbol	*exports;	/* Export symbol table */
};

/* CS_object: VM program */
typedef struct CS_entrypoint
{
	unsigned	position;	/* Start PC */
	unsigned	argv;		/* First register of argument list */
	unsigned	argc;		/* Number of arguments */
	int		argdefs[CS_MAXARGS];	/* Argument default values */
} CS_entrypoint;
struct CS_program
{
	CS_otypes	type;
	CS_bank		*bank;		/* Where the VM code is kept */
	CS_entrypoint	eps[CS_MAXEPS];	/* Entry points; 0 == "main" */
};

/* Oscillator */
struct CS_oscillator
{
	unsigned	phase;		/* Phase (16:16 fixp, 1.0/per) */
	unsigned	dphase;		/* Increment (16:16 fixp, 1.0/per) */
	CS_ramper	a;
	int		noise;		/* Current noise sample (S&H) */
	CS_wave		*wave;		/* Current waveform */
	CS_osccb	Process;	/* (Copy of wave->Process) */
};

/* Pan/volume mixer output stage */
struct CS_panmixer
{
	CS_ramper	vol;		/* Volume */
	CS_ramper	pan;		/* Horizontal pan position */
};

/* Voice and messages */
typedef enum CS_vstates
{
	CS_RUNNING = 0,
	CS_WAITING,	/* In a *DELAY instruction or similar state */
	CS_INTERRUPT,	/* Message handler context */
	CS_ENDING,	/* In top-level RETURN, ready for destruction */
	CS_FINALIZING	/* Detached, waiting for subvoices to finish */
} CS_vstates;

struct CS_stackentry
{
	CS_stackentry	*prev;
	CS_program	*program;
	CS_vstates	state;
	unsigned	pc;		/* PC of calling instruction */
	int		interrupt;	/* Restore timer and PC! */
	unsigned	timer;		/* Timer as message is handled */
	unsigned	firstreg;	/* First register saved */
	int		r[CS_REGISTERS - CS_CREGISTERS]; /* Saved registers */
};

typedef enum CS_messages
{
	/* Sub-voice messages */
	CSM_STARTSUB = 0,	/* Start new program on new sub-voice */
	CSM_SENDSUB,		/* Send message to sub-voices */
	CSM_DETACHSUB,		/* Detach sub-voice */
	
	/* Direct voice messages */
	CSM_MESSAGE,		/* Send message to voice */
	CSM_DETACH		/* Detach this voice */
} CS_messages;

typedef struct CS_messagebody
{
	unsigned	timestamp;	/* When to process (frames, 24:8 fixp) */
	uint16_t	addr;		/* Handle, program, ... */
	CS_vid		vid;
	CS_gid		gid;
	uint8_t		what;		/* Operation to perform */
	uint8_t		argc;		/* Argument count */
	int		a[CS_MAXARGS];	/* Arguments (16:16 fixp) */
} CS_messagebody;

struct CS_message
{
	CS_message	*next;
	CS_messagebody	m;
	NUMMSGS(unsigned number;)
};

struct CS_voice
{
	CS_vstates	state;		/* Current state */
	CS_voice	*next;		/* Next voice in list */
	CS_voice	*sub;		/* Sub-voices */
	CS_stackentry	*stack;		/* VM call stack */
	CS_message	*messages;	/* Message queue */
	CS_program	*program;
	int		id;		/* ID for voice addressing */
	unsigned	timer;		/* Delay timer (frames, 24:8 fixp) */
	unsigned	pc;		/* PC of calling instruction */
	int		r[CS_REGISTERS]; /* VM registers */
	CS_oscillator	osc;		/* Local oscillator */
	CS_panmixer	pmx;		/* Output stage */
	CS_voice	*sv[CS_REGISTERS];/* Attached sub-voices */
	CS_insert_cb	insert_cb;
	void		*insert_ud;
};

#define	CS_NEEDS_WORKBUFS(v)	(			\
		(v)->insert_cb ||			\
		((v)->pmx.vol.value != 0x1000000) ||	\
		((v)->pmx.vol.target != 0x1000000)	\
	)


/* Block - allocation unit for CS_stackentry and mixing buffers */
typedef union CS_block CS_block;
union CS_block
{
	CS_block	*next;
	CS_stackentry	stackentry;
	int		buffer[CS_MAXFRAG];
};


/* Compiler state */
struct CS_fixup
{
	CS_fixup	*next;
	unsigned	pos;		/* Position of instruction to fix */
};

typedef struct CS_lexstate
{
	int		codepos;	/* Source code read position */
	CS_tokens	token;
	int		val;		/* Value; int/fixp */
	CS_symbol	*sym;		/* Value; (new) symbol */
} CS_lexstate;

struct CS_cstate
{
	CS_state	*enginestate;	/* Parent engine state */
	CS_symbol	*symbols;	/* Symbol stack */
	const char	*code;		/* Source code buffer */
	CS_bank		*bank;		/* Target code bank */
	int		exportall;	/* Export everything from banks! */
	CS_lexstate	l, pl;		/* Current and previous lexer states */
	unsigned	regtop;		/* First free register */
	CS_program	*program;	/* Current program, if any */
	int		inhandler;	/* Disallow timing, RUN, SLEEP ,... */
	int		nocode;		/* Disallow code in current context  */
	CS_jumpbuf	jumpbuf;	/* Buffer for csc_Try()/csc_Throw() */
	CS_errors	error;		/* Error from csc_Throw() */
};

/* ChipSound state */
struct CS_state
{
	int		running;	/* State is initialized and running */
	int		flags;		/* 'flags' from cs_Open() */

/*FIXME: These should really be used with read/write barriers to be safe... */
	volatile unsigned now_frames;	/* Audio time of last cb (frames, 24:8) */
	volatile unsigned now_ticks;	/* Tick of last audio callback (ms) */
	volatile unsigned now_guard;	/* Guard, matching now_frames */

	unsigned	timestamp;	/* Current timestamp for async API */
	sfifo_t		fromapi;	/* Messages from async. API calls */

	CS_voice	*voicepool;	/* LIFO stack of voices */
	unsigned	totalvoices;	/* Number of voices in use + pool */
	unsigned	activevoices;	/* Number of voices in use */

	CS_block	*blockpool;	/* LIFO stack of memory blocks */
	CS_message	*messagepool;	/* LIFO stack of message containers */
	unsigned	now_fragstart;	/* For internal message timing */

	unsigned	maxobjects;	/* Size of 'objects' table (elements) */
	unsigned	ffobject;	/* First free object slot */
	CS_object	**objects;	/* Global table of objects */

	CS_symbol	*banks;		/* Symbol table for the banks */
	CS_voice	*groups[CS_MAXGIDS];	/* [0] is "root" */

	CS_program	*groupprog;	/* Voice program for groups */
	CS_program	*terminator;	/* Voice program for killed voices */
	uint32_t	rnd;
	unsigned	msdur;		/* One millisecond in samples (16:16) */
	SDL_AudioSpec	audiospec;

	unsigned	avgtick;	/* Timestamp for averaging start */
	unsigned	cputimesum;	/* Time spent in audio callback */
	unsigned	cputimecount;	/* Number of audio callbacks */
	unsigned	cputimemax;	/* Maximum time spent in audio cb */
	unsigned	cpuloadmax;	/* Maximum CPU load */
	unsigned	instructions;	/* VM instruction counter */

/*FIXME: Feedback delay unit... Should not be hardwired! */
	int		*delaybuf;
	int		delaypos;
	int		fbdelay;
	int		rdelay;
	int		ldelay;

	/* Simple output limiter from Audiality */
	int		*masterbuf;
	LIM_limiter	limiter;

#ifdef CS_ALSA_MIDI
	/* ALSA MIDI control */
	int		midiportid;
	snd_seq_t	*midiseqh;
	int		midiprg[16];
	int		midimod[16];
#endif
	CS_cstate	*c;		/* Compiler state */

	NUMMSGS(unsigned msgnum;)
};

/* Global state instance */
extern CS_state cs_globalstate;


/* Compiler open/close */
CS_errors cs_OpenCompiler(CS_state *st, int flags);
void cs_CloseCompiler(CS_cstate *c);

/*---------------------------------------------------------
	Object/handle management
---------------------------------------------------------*/

int cs_ObjAlloc(CS_state *st, CS_otypes type);
CS_object *cs_ObjGet(CS_state *st, unsigned handle, CS_otypes type);


/* Returns a pseudo random number in the range [0, 65535] */
static inline int cs_Noise(CS_state *st)
{
	st->rnd *= 1566083941UL;
	st->rnd++;
	return st->rnd * (st->rnd >> 16) >> 16;
}


/*---------------------------------------------------------
	Realtime block memory manager
---------------------------------------------------------*/

static inline CS_block *cs_NewBlock(void)
{
	CS_block *b = malloc(sizeof(CS_block));
	if(!b)
		return NULL;
#ifdef DEBUG
	fprintf(stderr, "ChipSound: Block pool exhausted!"
			"Allocated new block %p.\n", b);
#endif
	return b;
}

static inline CS_block *cs_AllocBlock(CS_state *st)
{
	CS_block *b = st->blockpool;
	if(b)
		st->blockpool = b->next;
	else
		b = cs_NewBlock();
	return b;
}

static inline void cs_FreeBlock(CS_state *st, void *block)
{
	((CS_block *)block)->next = st->blockpool;
	st->blockpool = (CS_block *)block;
}


/*---------------------------------------------------------
	Realtime message passing
---------------------------------------------------------*/

static inline CS_message *cs_NewMessage(void)
{
	CS_message *m = malloc(sizeof(CS_message));
	if(!m)
		return NULL;
#ifdef DEBUG
	fprintf(stderr, "ChipSound: Message pool exhausted!"
			"Allocated new message %p.\n", m);
#endif
	return m;
}

static inline CS_message *cs_AllocMessage(CS_state *st)
{
	CS_message *m = st->messagepool;
	if(m)
		st->messagepool = m->next;
	else
		m = cs_NewMessage();
	NUMMSGS(m->number = st->msgnum++;)
	return m;
}

static inline void cs_FreeMessage(CS_state *st, CS_message *m)
{
	m->next = st->messagepool;
	st->messagepool = m;
}

static inline void cs_SendMessage(CS_voice *v, CS_message *m)
{
	CS_message *pm = v->messages;
//printf("--- SendMessage(%p, %u) %f\t", v, m->number, m->m.timestamp / 256.0f);
	if(!pm || (cs_TSDiff(pm->m.timestamp, m->m.timestamp) > 0))
	{
		m->next = pm;
		v->messages = m;
	}
	else
	{
		while(pm->next && (cs_TSDiff(pm->next->m.timestamp,
				m->m.timestamp) <= 0))
			pm = pm->next;
		m->next = pm->next;
		pm->next = m;
	}
/*
	NUMMSGS(
		printf("\t[ ");
		pm = v->messages;
		while(pm)
		{
			printf("%u ", pm->number);
			pm = pm->next;
		}
		printf("]");
	)
*/
//printf("\n");
}


/*---------------------------------------------------------
	Oscillators
---------------------------------------------------------*/

void cso_Off(CS_oscillator *o, int *buffer, unsigned frames);
void cso_Noise(CS_oscillator *o, int *buffer, unsigned frames);
void cso_Wavetable(CS_oscillator *o, int *buffer, unsigned frames);

static inline void cs_OscWave(CS_oscillator *o, CS_state *st, unsigned wave)
{
	CS_wave *w = (CS_wave *)cs_ObjGet(st, wave, CS_TWAVE);
	o->wave = w;
	o->Process = w ? w->Process : cso_Off;
}


/*---------------------------------------------------------
	Voice management
---------------------------------------------------------*/

CS_voice *cs_VoiceNew(CS_state *st, CS_voice *parent);
CS_errors cs_VoiceCall(CS_state *st, CS_voice *v, CS_program *p, unsigned ep,
		int argc, int *argv);
void cs_VoiceKill(CS_state *st, CS_voice *v);
void cs_VoiceFree(CS_state *st, CS_voice **head);

static inline void cs_VoiceDetach(CS_voice *v)
{
	v->id = -1;
	if(v->state >= CS_ENDING)
		v->timer = 0;	/* Wake up and terminate! */
}


/*---------------------------------------------------------
	Async API message gateway
---------------------------------------------------------*/

/*
 * Set up buffers, sync constructs etc for the asynchronous API
 * 'buffer' is the playback time of the audio output buffer, in seconds.
 */
CS_errors cs_OpenAPI(CS_state *st, float buffer);

void cs_CloseAPI(CS_state *st);
CS_errors cs_PumpMessages(CS_state *st);


#endif /* CSINTERNALS_H */
