/*
 * ChipSound test program
 *
 * This code is in the public domain. Do what you like with it. NO WARRANTY!
 *
 * 2011-2012 David Olofson
 */

#include <stdlib.h>
#include <signal.h>
#include <math.h>
#include "SDL.h"
#include "chipsound.h"
#include "gui.h"

/* Display */
static SDL_Surface *screen = NULL;
static int screenw = 800;
static int screenh = 480;
static int screenbpp = 0;
static int screenflags = 0;

/* Oscilloscopes */
static int oscpos = 0;			/* Grab buffer position */
static int plotpos = 0;			/* Plot position */
static Sint32 *osc_left = NULL;		/* Left audio grab buffer */
static Sint32 *osc_right = NULL;	/* Right audio grab buffer */

/* Load info */
static Uint32 now = 0;
static Uint32 lastreset;

/* "System" */
static int samplerate = 44100;
static int audiobuf = 1024;
static int dbuffer = -1;		/* Sync delay buffer size */
static int csflags = CS_MIDI | CS_EXPORTALL;
static int do_exit = 0;

/* Playing and control */
static int sfx = 0;
static int oct = 0;
static int lastkey = 0;
static float mod = 0;
static int legato = 0;

static int demo = 0;
#define	DEMOC	5
static char *demov[DEMOC] = {
	"data/evilnoises.csl",
	"data/k2epilogue.csl",
	"data/k2intro.csl",
	"data/pulsetronic.csl",
	"data/k2trance.csl"
};
static char *songnames[DEMOC] = {
	"K2TitleFullSong",
	"PulseTronic",
	"K2Trance",
	"K2Epilogue",
	"EvilNoises"
};

/*-------------------------------------------------------------------
	Audio processing
-------------------------------------------------------------------*/

/* Grab data for the oscilloscopes */
static CS_errors grab_process(int **buffers, unsigned nbuffers, unsigned frames,
		void *userdata)
{
	int i;
	for(i = 0; i < frames; ++i)
		osc_left[(oscpos + i) % dbuffer] = buffers[0][i];
#if 0
	if(nbuffers >= 2)
		for(i = 0; i < frames; ++i)
			osc_right[(oscpos + i) % dbuffer] = buffers[1][i];
#else
	/* Stereo voices and groups not yet implemented in ChipSound! */
	for(i = 0; i < frames; ++i)
		osc_right[(oscpos + i) % dbuffer] = buffers[0][i];
#endif
	oscpos = (oscpos + frames) % dbuffer;
	plotpos = oscpos;
	return CS_OK;
}


/*-------------------------------------------------------------------
	File I/O
-------------------------------------------------------------------*/

static void load_sounds(int argc, char *argv[])
{
	int i, bank;
	if(demo)
	{
		argc = DEMOC + 1;
		argv = demov - 1;
	}
	cs_FreeBank(-1);
	cs_NewGroup(0, 1);
	for(i = 1; i < argc; ++i)
	{
		if(argv[i][0] == '-')
			continue;
		bank = cs_Load(argv[i]);
		if(bank < 0)
		{
			char buf[128];
			snprintf(buf, sizeof(buf), "Could not load \"%s\"!"
					" (%s)\n", argv[i],
					cs_ErrorString(-bank));
			gui_message(buf, -1);
		}
	}
}


static void select_object(void)
{
	char buf[128];
	snprintf(buf, sizeof(buf), "%s %d: %s\n",
			cs_TypeName(cs_GetType(cs_Get(-1, sfx))),
		 	sfx,
			cs_GetName(-1, sfx));
	gui_message(buf, -1);
}


static play_song(int song)
{
	char buf[128];
	snprintf(buf, sizeof(buf), "Playing song %s\n", songnames[song]);
	gui_message(buf, -1);
	cs_Start(1, 0, cs_Find(-1, songnames[song]), 0, 1, 0);
}


static void handle_events(int argc, char *argv[])
{
	char buf[128];
	SDL_Event event;
	while(SDL_PollEvent(&event))
	{
		switch(event.type)
		{
		  case SDL_QUIT:
			do_exit = 1;
			break;
		  case SDL_MOUSEBUTTONDOWN:
			break;
		  case SDL_MOUSEBUTTONUP:
			break;
		  case SDL_KEYDOWN:
			switch(event.key.keysym.sym)
			{
			  case SDLK_ESCAPE:
				do_exit = 1;
				break;
			  case SDLK_LCTRL:
			  case SDLK_RCTRL:
			  case SDLK_LSHIFT:
			  case SDLK_RSHIFT:
				legato = 1;
				break;
			  case SDLK_F1:
			  case SDLK_F2:
			  case SDLK_F3:
			  case SDLK_F4:
			  case SDLK_F5:
			  case SDLK_F6:
			  case SDLK_F7:
			  case SDLK_F8:
			  case SDLK_F9:
			  case SDLK_F10:
			  case SDLK_F11:
			  case SDLK_F12:
				if(legato)
				{
					cs_Send(1, 0, 2,
							(event.key.keysym.sym -
							SDLK_F1) / 12.0f + oct);
					cs_Send(1, 0, 1, 1);
				}
				else
					cs_Start(1, event.key.keysym.sym,
					 		cs_Get(-1, sfx),
							(event.key.keysym.sym -
							SDLK_F1) / 12.0f + oct,
							1, mod);
				lastkey = event.key.keysym.sym;
				break;
			  case SDLK_1:
			  case SDLK_2:
			  case SDLK_3:
			  case SDLK_4:
			  case SDLK_5:
				play_song(event.key.keysym.sym - SDLK_1);
				break;
			  case SDLK_KP1:
			  case SDLK_KP2:
			  case SDLK_KP3:
			  case SDLK_KP4:
			  case SDLK_KP5:
				play_song(event.key.keysym.sym - SDLK_KP1);
				break;
			  case SDLK_p:
				cs_Start(1, 0, cs_Get(-1, sfx), oct, 1, mod);
				break;
			  case SDLK_y:
				cs_Kill(1, 0);
				cs_Kill(CS_MAXGIDS - 1, 0); /* Built-in MIDI */
				break;
			  case SDLK_r:
				load_sounds(argc, argv);
				break;
			  case SDLK_PLUS:
			  case SDLK_KP_PLUS:
				++sfx;
				select_object();
				break;
			  case SDLK_MINUS:
			  case SDLK_KP_MINUS:
				--sfx;
				select_object();
				break;
			  case SDLK_PAGEUP:
				++oct;
				snprintf(buf, sizeof(buf), "Octave: %d\n", oct);
				gui_message(buf, -1);
				break;
			  case SDLK_PAGEDOWN:
				--oct;
				snprintf(buf, sizeof(buf), "Octave: %d\n", oct);
				gui_message(buf, -1);
				break;
			  case SDLK_KP_MULTIPLY:
				mod += .1;
				cs_Send(1, 0, 3, mod);
				snprintf(buf, sizeof(buf), "Modulation: %f\n", mod);
				gui_message(buf, -1);
				break;
			  case SDLK_KP_DIVIDE:
				mod -= .1;
				cs_Send(1, 0, 3, mod);
				snprintf(buf, sizeof(buf), "Modulation: %f\n", mod);
				gui_message(buf, -1);
				break;
			  default:
				break;
			}
			break;
		  case SDL_KEYUP:
			switch(event.key.keysym.sym)
			{
			  case SDLK_LCTRL:
			  case SDLK_RCTRL:
			  case SDLK_LSHIFT:
			  case SDLK_RSHIFT:
				legato = 0;
				cs_Send(1, 0, 1);
				cs_Detach(1, 0);
				break;
			  case SDLK_F1:
			  case SDLK_F2:
			  case SDLK_F3:
			  case SDLK_F4:
			  case SDLK_F5:
			  case SDLK_F6:
			  case SDLK_F7:
			  case SDLK_F8:
			  case SDLK_F9:
			  case SDLK_F10:
			  case SDLK_F11:
			  case SDLK_F12:
				if(legato)
				{
					if(event.key.keysym.sym == lastkey)
						cs_Send(1, 0, 1);
				}
				else
				{
					cs_Send(1, event.key.keysym.sym, 1);
					cs_Detach(1, event.key.keysym.sym);
				}
				break;
			  default:
				break;
			}
			break;
		  default:
			break;
		}
	}
}


static void update_main(int dt)
{
	int w = (screen->w - 240) / 2 - 8;
	gui_oscilloscope(osc_left, dbuffer, plotpos,
			240, 8, w, 128, screen);
	gui_oscilloscope(osc_right, dbuffer, plotpos,
			240 + w + 8, 8, w, 128, screen);
	gui_cpuload(cs_GetGlobalProp(CS_PCPULOADAVG));
	gui_voices(cs_GetGlobalProp(CS_PACTIVEVOICES));
	if(now - lastreset > 300)
	{
		gui_instructions(cs_GetGlobalProp(CS_PINSTRUCTIONS) * 1000 /
				(now - lastreset));
		cs_SetGlobalProp(CS_PCPULOADAVG, 0);
		cs_SetGlobalProp(CS_PINSTRUCTIONS, 0);
		lastreset = now;
	}
}


static void parse_args(int argc, char *argv[])
{
	int i;
	demo = 1;
	for(i = 1; i < argc; ++i)
	{
		if(argv[i][0] != '-')
		{
			demo = 0;	/* Filename specified - no demo! */
			continue;
		}
		if(strncmp(argv[i], "-f", 2) == 0)
		{
			screenflags |= SDL_FULLSCREEN;
			printf("[Fullscreen mode]\n");
		}
		else if(strncmp(argv[i], "-w", 2) == 0)
		{
			screenw = atoi(&argv[i][2]);
			printf("[Display width: %d]\n", screenw);
		}
		else if(strncmp(argv[i], "-h", 2) == 0)
		{
			screenh = atoi(&argv[i][2]);
			printf("[Display height: %d]\n", screenh);
		}
		else if(strncmp(argv[i], "-b", 2) == 0)
		{
			screenbpp = atoi(&argv[i][2]);
			printf("[Display bpp: %d]\n", screenbpp);
		}
		else if(strncmp(argv[i], "-a", 2) == 0)
		{
			audiobuf = atoi(&argv[i][2]);
			printf("[Audio buffer: %d]\n", audiobuf);
		}
		else if(strncmp(argv[i], "-r", 2) == 0)
		{
			samplerate = atoi(&argv[i][2]);
			printf("[Audio sample rate: %d]\n", samplerate);
		}
		else
		{
			fprintf(stderr, "Unknown switch '%s'!\n", argv[i]);
			exit(1);
		}
	}
}


static void breakhandler(int a)
{
	do_exit = 1;
}


int main(int argc, char *argv[])
{
	int lasttick;
	parse_args(argc, argv);
	if(SDL_Init(SDL_INIT_VIDEO) < 0)
		exit(1);
	atexit(SDL_Quit);
	signal(SIGTERM, breakhandler);
	signal(SIGINT, breakhandler);

	screen = SDL_SetVideoMode(screenw, screenh, screenbpp, screenflags);
	SDL_WM_SetCaption("ChipSound Test Program", "ChipSound");
	SDL_ShowCursor(1);
	if(gui_open(screen) < 0)
		exit(1);

	if(cs_Open(samplerate, audiobuf, 2, csflags))
		exit(1);
	if(dbuffer < 0)
		dbuffer = audiobuf * 3;
	osc_left = calloc(dbuffer, sizeof(Sint32));
	osc_right = calloc(dbuffer, sizeof(Sint32));
	if(!osc_left || !osc_right)
	{
		fprintf(stderr, "Couldn't allocate visualization buffers!\n");
		exit(1);
	}

	cs_Insert(0, grab_process, NULL);

	load_sounds(argc, argv);

	gui_draw_screen(demo);
	
	lastreset = lasttick = SDL_GetTicks();
	while(!do_exit)
	{
		int tick = SDL_GetTicks();
		int dt = tick - lasttick;
		now += dt;
		lasttick = tick;
		handle_events(argc, argv);
		update_main(dt);
		gui_refresh();
		SDL_Delay(10);
	}

	cs_Close();
	gui_close();
	SDL_Quit();
	return 0;
}
